# Copyright (C) 2017 Red Hat, Inc., Bryn M. Reeves <bmr@redhat.com>
#
# command_tests.py - Boom command API tests.
#
# This file is part of the boom project.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions
# of the GNU General Public License v.2.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
import unittest
import logging
from sys import stdout
from os import listdir, makedirs
from os.path import abspath, basename, dirname, exists, join
from glob import glob
import shutil
import re

# Python3 moves StringIO to io
try:
    from StringIO import StringIO
except:
    from io import StringIO

log = logging.getLogger()
log.level = logging.DEBUG
log.addHandler(logging.FileHandler("test.log"))

from boom import *
from boom.osprofile import *
from boom.bootloader import *
from boom.hostprofile import *
from boom.command import *
from boom.config import *
from boom.report import *

# For access to non-exported members
import boom.command

from tests import *

BOOT_ROOT_TEST = abspath("./tests")
config = BoomConfig()
config.legacy_enable = False
config.legacy_sync = False
set_boom_config(config)
set_boot_path(BOOT_ROOT_TEST)

debug_masks = ['profile', 'entry', 'report', 'command', 'all']


class CommandHelperTests(unittest.TestCase):
    """Test internal boom.command helpers: methods in this part of the
        test suite import boom.command directly in order to access the
        non-public helper routines not included in __all__.
    """
    def test_int_if_val_with_val(self):
        import boom.command
        val = "1"
        self.assertEqual(boom.command._int_if_val(val), int(val))

    def test_int_if_val_with_none(self):
        import boom.command
        val = None
        self.assertEqual(boom.command._int_if_val(val), None)

    def test_int_if_val_with_badint(self):
        import boom.command
        val = "qux"
        with self.assertRaises(ValueError) as cm:
            boom.command._int_if_val(val)

    def test_subvol_from_arg_subvol(self):
        import boom.command
        xtuple = ("/svol", None)
        self.assertEqual(boom.command._subvol_from_arg("/svol"), xtuple)

    def test_subvol_from_arg_subvolid(self):
        import boom.command
        xtuple = (None, "23")
        self.assertEqual(boom.command._subvol_from_arg("23"), xtuple)

    def test_subvol_from_arg_none(self):
        import boom.command
        self.assertEqual(boom.command._subvol_from_arg(None), (None, None))

    def test_str_indent(self):
        import boom.command
        instr = "1\n2\n3\n4"
        xstr = "    1\n    2\n    3\n    4"
        indent = 4
        outstr = boom.command._str_indent(instr, indent)
        self.assertEqual(outstr, xstr)

    def test_str_indent_bad_indent(self):
        import boom.command
        instr = "1\n2\n3\n4"
        indent = "qux"
        with self.assertRaises(TypeError) as cm:
            outstr = boom.command._str_indent(instr, indent)

    def test_str_indent_bad_str(self):
        import boom.command
        instr = None
        indent = 4
        with self.assertRaises(AttributeError) as cm:
            outstr = boom.command._str_indent(instr, indent)

    def test_canonicalize_lv_name(self):
        import boom.command
        xlv = "vg/lv"
        for lvstr in  ["vg/lv", "/dev/vg/lv"]:
            self.assertEqual(xlv, boom.command._canonicalize_lv_name(lvstr))

    def test_canonicalize_lv_name_bad_lv(self):
        import boom.command
        with self.assertRaises(ValueError) as cm:
            boom.command._canonicalize_lv_name("vg/lv/foo/bar/baz")
        with self.assertRaises(ValueError) as cm:
            boom.command._canonicalize_lv_name("vg-lv")
        with self.assertRaises(ValueError) as cm:
            boom.command._canonicalize_lv_name("/dev/mapper/vg-lv")

    def test_expand_fields_defaults(self):
        import boom.command
        default = "f1,f2,f3"
        xfield = default
        self.assertEqual(xfield, boom.command._expand_fields(default, ""))

    def test_expand_fields_replace(self):
        import boom.command
        default = "f1,f2,f3"
        options = "f4,f5,f6"
        xfield = options
        self.assertEqual(xfield, boom.command._expand_fields(default, options))

    def test_expand_fields_add(self):
        import boom.command
        default = "f1,f2,f3"
        options = "+f4,f5,f6"
        xfield = default + ',' + options[1:]
        self.assertEqual(xfield, boom.command._expand_fields(default, options))

    def test_set_debug_no_debug_arg(self):
        """Test set_debug() with an empty debug mask argument.
        """
        import boom.command
        boom.command.set_debug(None)

    def test_set_debug_args_one(self):
        """Test set_debug() with a single debug mask argument.
        """
        import boom.command
        for mask in debug_masks:
            boom.command.set_debug(mask)

    def test_set_debug_args_all(self):
        """Test set_debug() with a list of debug mask arguments.
        """
        import boom.command
        all_masks = ",".join(debug_masks[:-1])
        boom.command.set_debug(all_masks)

    def test_set_debug_no_debug_arg(self):
        """Test set_debug() with a bad debug mask argument.
        """
        import boom.command
        with self.assertRaises(ValueError) as cm:
            boom.command.set_debug("nosuchmask")

    def test_setup_logging(self):
        """Test the setup_logging() command helper.
        """
        import boom.command
        args = MockArgs()
        boom.command.setup_logging(args)

    @unittest.skipIf(not have_grub1(), "requires grub1")
    def test_show_legacy_default(self):
        """Test the show_legacy() command helper.
        """
        import boom.command
        boom.command.show_legacy()

    def test__get_machine_id(self):
        # FIXME: does not cover _DBUS_MACHINE_ID hosts or exceptions
        # reading /etc/machine-id.
        machine_id = boom.command._get_machine_id()
        self.assertTrue(machine_id)

    def test__merge_add_del_opts_no_op(self):
        bp = BootParams(version="1.1.1", root_device="/dev/vg00/lvol0")
        _merge_add_del_opts = boom.command._merge_add_del_opts

        # Empty bp and empty add/del lead to empty results
        to_add = ""
        to_del = ""
        (add_opts, del_opts) = _merge_add_del_opts(bp, to_add, to_del)
        self.assertEqual([], add_opts)
        self.assertEqual([], del_opts)

        # Existing add/del with no modifications
        bp.add_opts = ["debug"]
        bp.del_opts = ["rhgb quiet"]
        (add_opts, del_opts) = _merge_add_del_opts(bp, to_add, to_del)
        self.assertEqual(bp.add_opts, add_opts)
        self.assertEqual(bp.del_opts, del_opts)

    def test__merge_add_del_opts_with_add(self):
        bp = BootParams(version="1.1.1", root_device="/dev/vg00/lvol0")
        _merge_add_del_opts = boom.command._merge_add_del_opts

        # Empty bp and add places add in final add_opts
        to_add = "debug"
        to_del = ""
        (add_opts, del_opts) = _merge_add_del_opts(bp, to_add, to_del)
        self.assertEqual([to_add], add_opts)
        self.assertEqual([], del_opts)

    def test__merge_add_del_opts_with_del(self):
        bp = BootParams(version="1.1.1", root_device="/dev/vg00/lvol0")
        _merge_add_del_opts = boom.command._merge_add_del_opts

        # Empty bp and del places del in final del_opts
        to_add = ""
        to_del = "rhgb quiet"
        (add_opts, del_opts) = _merge_add_del_opts(bp, to_add, to_del)
        self.assertEqual([], add_opts)
        self.assertEqual(to_del.split(), del_opts)

    def test__merge_add_del_opts_with_add_appends(self):
        bp = BootParams(version="1.1.1", root_device="/dev/vg00/lvol0",
                        add_opts=["log_buf_len=16M"])
        _merge_add_del_opts = boom.command._merge_add_del_opts

        # bp with add_opts and and command line add gives correct order
        # of final add_opts list (command line additions appended).
        to_add = "debug"
        to_del = ""
        (add_opts, del_opts) = _merge_add_del_opts(bp, to_add, to_del)
        self.assertEqual(["log_buf_len=16M", "debug"], add_opts)


# Default test OsProfile identifiers
test_os_id = "9cb53ddda889d6285fd9ab985a4c47025884999f"
test_os_disp_id = test_os_id[0:6]

test_lv = get_logical_volume()
test_root_lv = get_root_lv()

def get_create_cmd_args():
    """Return a correct MockArgs object for a call to the _create_cmd()
        helper. Tests that should fail modify the fields returned to
        generate the required error.
    """
    args = MockArgs()
    args.profile = test_os_disp_id
    args.title = "ATITLE"
    args.version = "2.6.0"
    args.machine_id = "ffffffff"
    args.root_device = get_logical_volume()
    args.root_lv = get_root_lv()
    return args


class CommandTests(unittest.TestCase):
    """Test boom.command APIs
    """

    # Master BLS loader directory for sandbox
    loader_path = join(BOOT_ROOT_TEST, "loader")

    # Master boom configuration path for sandbox
    boom_path = join(BOOT_ROOT_TEST, "boom")

    # Master grub configuration path for sandbox
    grub_path = join(BOOT_ROOT_TEST, "grub")

    # Test fixture init/cleanup
    def setUp(self):
        """Set up a test fixture for the CommandTests class.

            Defines standard objects for use in these tests.
        """
        reset_sandbox()

        # Sandbox paths
        boot_sandbox = join(SANDBOX_PATH, "boot")
        boom_sandbox = join(SANDBOX_PATH, "boot/boom")
        grub_sandbox = join(SANDBOX_PATH, "boot/grub")
        loader_sandbox = join(SANDBOX_PATH, "boot/loader")

        # Initialise sandbox from master
        makedirs(boot_sandbox)
        shutil.copytree(self.boom_path, boom_sandbox)
        shutil.copytree(self.loader_path, loader_sandbox)
        shutil.copytree(self.grub_path, grub_sandbox)

        # Copy boot images
        images = glob(join(BOOT_ROOT_TEST, "initramfs*"))
        images += glob(join(BOOT_ROOT_TEST, "vmlinuz*"))
        for image in images:
            def _dotfile(img_path):
                pattern = ".%s.boomrestored"
                img_name = basename(img_path)
                img_dir = dirname(img_path)
                return join(img_dir, pattern % img_name)

            shutil.copy2(image, boot_sandbox)
            if exists(_dotfile(image)):
                shutil.copy2(_dotfile(image), boot_sandbox)

        # Set boom paths
        set_boot_path(boot_sandbox)

        # Tests that deal with legacy configs will enable this.
        config = BoomConfig()
        config.legacy_enable = False
        config.legacy_sync = False

        # Reset profiles, entries, and host profiles to known state.
        load_profiles()
        load_entries()
        load_host_profiles()

    def tearDown(self):
        # Drop any in-memory entries and profiles modified by tests
        drop_entries()
        drop_profiles()
        drop_host_profiles()

        # Clear sandbox data
        rm_sandbox()
        reset_boom_paths()

    def test_command_find_profile_with_profile_arg(self):
        import boom.command
        _find_profile = boom.command._find_profile
        cmd_args = MockArgs()
        cmd_args.profile = "d4439b7d2f928c39f1160c0b0291407e5990b9e0" # F26
        cmd_args.machine_id = "12345" # No HostProfile
        osp = _find_profile(cmd_args, "", cmd_args.machine_id, "test")
        self.assertEqual(osp.os_id, cmd_args.profile)

    def test_command_find_profile_with_version_arg(self):
        import boom.command
        _find_profile = boom.command._find_profile
        cmd_args = MockArgs()
        cmd_args.profile = None
        cmd_args.version = "4.16.11-100.fc26.x86_64" # F26
        cmd_args.machine_id = "12345" # No HostProfile
        xprofile = "d4439b7d2f928c39f1160c0b0291407e5990b9e0"
        osp = _find_profile(cmd_args, cmd_args.version,
                            cmd_args.machine_id, "test")
        self.assertEqual(osp.os_id, xprofile)

    def test_command_find_profile_with_bad_version_arg(self):
        import boom.command
        _find_profile = boom.command._find_profile
        cmd_args = MockArgs()
        cmd_args.profile = None
        cmd_args.version = "4.16.11-100.x86_64" # no match
        cmd_args.machine_id = "12345" # No HostProfile
        xprofile = "d4439b7d2f928c39f1160c0b0291407e5990b9e0"
        osp = _find_profile(cmd_args, "", cmd_args.machine_id, "test")
        self.assertEqual(osp, None)

    def test_command_find_profile_bad_profile(self):
        import boom.command
        _find_profile = boom.command._find_profile
        cmd_args = MockArgs()
        cmd_args.profile = "quxquxquxquxquxquxquxqux" # nonexistent
        cmd_args.machine_id = "12345" # No HostProfile
        osp = _find_profile(cmd_args, "", cmd_args.machine_id, "test")
        self.assertEqual(osp, None)

    def test_command_find_profile_ambiguous_profile(self):
        import boom.command
        _find_profile = boom.command._find_profile
        cmd_args = MockArgs()
        cmd_args.profile = "9" # ambiguous
        cmd_args.machine_id = "12345" # No HostProfile
        osp = _find_profile(cmd_args, "", cmd_args.machine_id, "test")
        self.assertEqual(osp, None)

    def test_command_find_profile_ambiguous_host(self):
        import boom.command
        _find_profile = boom.command._find_profile
        cmd_args = MockArgs()
        cmd_args.profile = ""
        cmd_args.machine_id = "fffffffffff" # Ambiguous HostProfile
        osp = _find_profile(cmd_args, "", cmd_args.machine_id, "test")
        self.assertEqual(osp, None)

    def test_command_find_profile_host(self):
        import boom.command
        _find_profile = boom.command._find_profile
        cmd_args = MockArgs()
        cmd_args.profile = ""
        cmd_args.machine_id = "ffffffffffffc"
        cmd_args.label = ""
        hp = _find_profile(cmd_args, "", cmd_args.machine_id, "test")
        self.assertTrue(hp)
        self.assertTrue(hasattr(hp, "add_opts"))

    def test_command_find_profile_host_os_mismatch(self):
        import boom.command
        _find_profile = boom.command._find_profile
        cmd_args = MockArgs()
        cmd_args.profile = "3fc389bba581e5b20c6a46c7fc31b04be465e973"
        cmd_args.machine_id = "ffffffffffffc"
        cmd_args.label = ""
        hp = _find_profile(cmd_args, "", cmd_args.machine_id, "test")
        self.assertFalse(hp)

    def test_command_find_profile_no_matching(self):
        import boom.command
        _find_profile = boom.command._find_profile
        cmd_args = MockArgs()
        cmd_args.profile = ""
        cmd_args.machine_id = "1111111111111111" # no matching
        hp = _find_profile(cmd_args, "", cmd_args.machine_id,
                           "test", optional=False)
        self.assertFalse(hp)

    #
    # API call tests
    #
    # BootEntry tests
    #

    def test_list_entries(self):
        path = boom_entries_path()
        nr = len([p for p in listdir(path) if p.endswith(".conf")])
        bes = list_entries()
        self.assertTrue(len(bes), nr)

    def test_list_entries_match_machine_id(self):
        machine_id = "611f38fd887d41dea7eb3403b2730a76"
        path = boom_entries_path()
        nr = len([p for p in listdir(path) if p.startswith(machine_id)])
        bes = list_entries(Selection(machine_id=machine_id))
        self.assertTrue(len(bes), nr)

    def test_list_entries_match_version(self):
        version = "4.10.17-100.fc24.x86_64"
        path = boom_entries_path()
        nr = len([p for p in listdir(path) if version in p])
        bes = list_entries(Selection(version=version))
        self.assertEqual(len(bes), nr)

    @unittest.skipIf(not have_root_lv(), "requires root LV")
    def test_create_entry_notitle(self):
        # Fedora 24 (Workstation Edition)
        osp = get_os_profile_by_id(test_os_id)
        osp.title = None
        with self.assertRaises(ValueError) as cm:
            be = create_entry(None, "2.6.0", "ffffffff", test_lv,
                              lvm_root_lv=test_root_lv, profile=osp)

    @unittest.skipIf(not have_root_lv(), "requires root LV")
    def test_create_entry_noversion(self):
        # Fedora 24 (Workstation Edition)
        osp = get_os_profile_by_id(test_os_id)
        with self.assertRaises(ValueError) as cm:
            be = create_entry("ATITLE", None, "ffffffff", test_lv,
                              lvm_root_lv=test_root_lv, profile=osp)

    @unittest.skipIf(not have_root_lv(), "requires root LV")
    def test_create_entry_nomachineid(self):
        # Fedora 24 (Workstation Edition)
        osp = get_os_profile_by_id(test_os_id)
        with self.assertRaises(ValueError) as cm:
            be = create_entry("ATITLE", "2.6.0", "", test_lv,
                              lvm_root_lv=test_root_lv, profile=osp)

    @unittest.skipIf(not have_root_lv(), "requires root LV")
    def test_create_entry_norootdevice(self):
        # FIXME: should this default from the lvm_root_lv?
        # Fedora 24 (Workstation Edition)
        osp = get_os_profile_by_id(test_os_id)
        with self.assertRaises(ValueError) as cm:
            be = create_entry("ATITLE", "2.6.0", "ffffffff", None,
                              lvm_root_lv=test_root_lv, profile=osp)

    @unittest.skipIf(not have_root_lv(), "requires root LV")
    def test_create_entry_noosprofile(self):
        # Fedora 24 (Workstation Edition)
        osp = get_os_profile_by_id(test_os_id)
        with self.assertRaises(ValueError) as cm:
            be = create_entry("ATITLE", "2.6.0", "ffffffff",
                              test_lv, lvm_root_lv=test_root_lv)

    def test_create_dupe(self):
        # Fedora 24 (Workstation Edition)
        osp = get_os_profile_by_id(test_os_id)

        title = "Fedora (4.1.1-100.fc24.x86_64) 24 (Workstation Edition)"
        machine_id = "611f38fd887d41dea7eb3403b2730a76"
        version = "4.1.1-100.fc24"
        root_device = "/dev/sda5"
        btrfs_subvol_id = "23"

        with self.assertRaises(ValueError) as cm:
            create_entry(title, version, machine_id, root_device,
                         btrfs_subvol_id=btrfs_subvol_id, profile=osp,
                         allow_no_dev=True)

    @unittest.skipIf(not have_root_lv(), "requires root LV")
    def test_create_delete_entry(self):
        # Fedora 24 (Workstation Edition)
        osp = get_os_profile_by_id(test_os_id)
        be = create_entry("ATITLE", "2.6.0", "ffffffff", test_lv,
                          lvm_root_lv=test_root_lv, profile=osp)
        self.assertTrue(exists(be._entry_path))

        delete_entries(Selection(boot_id=be.boot_id))
        self.assertFalse(exists(be._entry_path))

    @unittest.skipIf(not have_grub1() or not have_root_lv(), "requires "
                     "grub1 and LVM")
    def test_create_delete_entry_with_legacy(self):
        config = BoomConfig()
        config.legacy_enable = True
        config.legacy_sync = True
        set_boom_config(config)
        set_boot_path(BOOT_ROOT_TEST)

        # Fedora 24 (Workstation Edition)
        osp = get_os_profile_by_id(test_os_id)
        be = create_entry("ATITLE", "2.6.0", "ffffffff", test_lv,
                          lvm_root_lv=test_root_lv, profile=osp)
        self.assertTrue(exists(be._entry_path))

        delete_entries(Selection(boot_id=be.boot_id))
        self.assertFalse(exists(be._entry_path))


    def test_delete_entries_no_matching_raises(self):
        with self.assertRaises(IndexError) as cm:
            delete_entries(Selection(boot_id="thereisnospoon"))

    def test_clone_entry_no_boot_id(self):
        with self.assertRaises(ValueError) as cm:
            bad_be = clone_entry(Selection())

    def test_clone_entry_no_matching_boot_id(self):
        with self.assertRaises(ValueError) as cm:
            bad_be = clone_entry(Selection(boot_id="qqqqqqq"), title="FAIL")

    def test_clone_entry_ambiguous_boot_id(self):
        with self.assertRaises(ValueError) as cm:
            bad_be = clone_entry(Selection(boot_id="6"), title="NEWTITLE")


    def test_clone_entry_add_opts(self):
        be = clone_entry(Selection(boot_id="9591d36"), title="NEWNEWTITLE",
                         add_opts="foo", allow_no_dev=True)
        self.assertTrue(exists(be._entry_path))
        be.delete_entry()
        self.assertFalse(exists(be._entry_path))

    def test_clone_entry_del_opts(self):
        be = clone_entry(Selection(boot_id="9591d36"), title="NEWNEWTITLE",
                         del_opts="rhgb quiet", allow_no_dev=True)
        self.assertTrue(exists(be._entry_path))
        be.delete_entry()
        self.assertFalse(exists(be._entry_path))

    @unittest.skipIf(not have_root_lv(), "requires root LV")
    def test_clone_entry_del_opts_and_re_add(self):
        # Fedora 24 (Workstation Edition)

        # Delete rhgb quiet
        osp = get_os_profile_by_id(test_os_id)
        be = create_entry("delopts", "2.6.0", "ffffffff", test_lv,
                          lvm_root_lv=test_root_lv, profile=osp,
                          del_opts="rhgb quiet")

        # Assert it's gone
        self.assertFalse("rhgb quiet" in be.options)

        be2 = clone_entry(Selection(boot_id=be.boot_id), title="addoptsclone",
                          add_opts="rhgb quiet")

        # Assert it's back
        self.assertTrue("rhgb quiet" in be2.options)

    @unittest.skipIf(not have_root_lv(), "requires root LV")
    def test_clone_entry_add_opts_and_re_del(self):
        # Fedora 24 (Workstation Edition)

        # Add debug
        osp = get_os_profile_by_id(test_os_id)
        be = create_entry("addopts", "2.6.0", "ffffffff", test_lv,
                          lvm_root_lv=test_root_lv, profile=osp,
                          add_opts="debug")

        # Assert it's there
        self.assertTrue("debug" in be.options)

        be2 = clone_entry(Selection(boot_id=be.boot_id), title="deloptsclone",
                          del_opts="debug")

        # Assert it's gone
        self.assertFalse("debug" in be2.options)

    @unittest.skipIf(not have_root_lv(), "requires root LV")
    def test_clone_delete_entry(self):
        # Fedora 24 (Workstation Edition)
        osp = get_os_profile_by_id(test_os_id)
        be = create_entry("ATITLE", "2.6.0", "ffffffff", test_lv,
                          lvm_root_lv=test_root_lv, profile=osp)
        self.assertTrue(exists(be._entry_path))

        be2 = clone_entry(Selection(boot_id=be.boot_id), title="ANEWTITLE",
                          version="2.6.1")

        self.assertTrue(exists(be2._entry_path))

        be.delete_entry()
        be2.delete_entry()

        self.assertFalse(exists(be._entry_path))
        self.assertFalse(exists(be2._entry_path))

    @unittest.skipIf(not have_root_lv(), "requires root LV")
    def test_clone_entry_no_args(self):
        # Fedora 24 (Workstation Edition)
        osp = get_os_profile_by_id(test_os_id)
        be = create_entry("ATITLE", "2.6.0", "ffffffff", test_lv,
                          lvm_root_lv=test_root_lv, profile=osp)
        self.assertTrue(exists(be._entry_path))

        with self.assertRaises(ValueError) as cm:
            be2 = clone_entry(Selection(boot_id=be.boot_id))

        be.delete_entry()

    def test_clone_entry_with_add_del_opts(self):
        # Entry with options +"debug" -"rhgb quiet"
        orig_boot_id = "78861b7"
        # Use allow_no_dev=True here since we are cloning an existing
        # entry on a system with unknown devices.
        be = clone_entry(Selection(boot_id=orig_boot_id),
                         title="clone with addopts", allow_no_dev=True)
        orig_be = find_entries(Selection(boot_id=orig_boot_id))[0]
        self.assertTrue(orig_be)
        self.assertTrue(be)
        self.assertEqual(orig_be.options, be.options)
        be.delete_entry()

    @unittest.skipIf(not have_root_lv(), "requires root LV")
    def test_clone_dupe(self):
        # Fedora 24 (Workstation Edition)
        osp = get_os_profile_by_id(test_os_id)
        be = create_entry("CLONE_TEST", "2.6.0", "ffffffff", test_lv,
                          lvm_root_lv=test_root_lv, profile=osp)
        self.assertTrue(exists(be._entry_path))

        be2 = clone_entry(Selection(boot_id=be.boot_id), title="ANEWTITLE",
                          version="2.6.1")

        with self.assertRaises(ValueError) as cm:
            be3 = clone_entry(Selection(boot_id=be.boot_id), title="ANEWTITLE",
                              version="2.6.1")

        be.delete_entry()
        be2.delete_entry()

    def test_edit_entry_no_boot_id(self):
        with self.assertRaises(ValueError) as cm:
            bad_be = edit_entry(Selection())

    def test_edit_entry_no_matching_boot_id(self):
        with self.assertRaises(ValueError) as cm:
            bad_be = edit_entry(Selection(boot_id="qqqqqqq"), title="FAIL")

    def test_edit_entry_ambiguous_boot_id(self):
        with self.assertRaises(ValueError) as cm:
            bad_be = edit_entry(Selection(boot_id="6"), title="NEWTITLE")


    @unittest.skipIf(not have_root_lv(), "requires root LV")
    def test_edit_entry_add_opts(self):
        # Fedora 24 (Workstation Edition)
        osp = get_os_profile_by_id(test_os_id)
        orig_be = create_entry("EDIT_TEST", "2.6.0", "ffffffff",
                               test_lv, lvm_root_lv=test_root_lv,
                               profile=osp)

        # Confirm original entry has been written
        self.assertTrue(exists(orig_be._entry_path))

        # Save these - they will be overwritten by the edit operation
        orig_id = orig_be.boot_id
        orig_entry_path = orig_be._entry_path

        edit_title = "EDITED_TITLE"
        edit_add_opts = "foo"

        # FIXME: restore allow_no_dev
        edit_be = edit_entry(Selection(boot_id=orig_id), title=edit_title,
                              add_opts=edit_add_opts)

        # Confirm edited entry has been written
        self.assertTrue(exists(edit_be._entry_path))

        # Confirm original entry has been removed
        self.assertFalse(exists(orig_entry_path))

        # Verify new boot_id
        self.assertFalse(orig_id == edit_be.boot_id)

        # Verify edited title and options
        self.assertEqual(edit_title, edit_be.title)
        self.assertEqual(edit_be.bp.add_opts, [edit_add_opts])
        self.assertTrue(edit_add_opts in edit_be.options)

        # Clean up entries
        edit_be.delete_entry()

    @unittest.skipIf(not have_root_lv(), "requires root LV")
    def test_edit_entry_add_opts_with_add_opts(self):
        edit_title = "EDITED_TITLE"
        edit_add_opts = "foo"
        orig_add_opts = "bar"

        # Fedora 24 (Workstation Edition)
        osp = get_os_profile_by_id(test_os_id)
        orig_be = create_entry("EDIT_TEST", "2.6.0", "ffffffff",
                               test_lv, lvm_root_lv=test_root_lv,
                               add_opts="bar", profile=osp)

        # Confirm original entry has been written
        self.assertTrue(exists(orig_be._entry_path))

        # Save these - they will be overwritten by the edit operation
        orig_id = orig_be.boot_id
        orig_entry_path = orig_be._entry_path

        # FIXME: restore allow_no_dev
        edit_be = edit_entry(Selection(boot_id=orig_id), title=edit_title,
                             add_opts=edit_add_opts)

        # Confirm edited entry has been written
        self.assertTrue(exists(edit_be._entry_path))

        # Confirm original entry has been removed
        self.assertFalse(exists(orig_entry_path))

        # Verify new boot_id
        self.assertFalse(orig_id == edit_be.boot_id)

        # Verify edited title and options
        self.assertEqual(edit_title, edit_be.title)

        # Sort the opts lists as Python3 does not guarantee ordering
        sorted_bp_add_opts = sorted(edit_be.bp.add_opts)
        sorted_edit_and_orig_opts = sorted([edit_add_opts, orig_add_opts])
        self.assertEqual(sorted_bp_add_opts, sorted_edit_and_orig_opts)

        # Verify original added opts
        self.assertTrue(orig_add_opts in edit_be.options)
        # Verify edit added opts
        self.assertTrue(edit_add_opts in edit_be.options)

        # Clean up entries
        edit_be.delete_entry()

    @unittest.skipIf(not have_root_lv(), "requires root LV")
    def test_edit_entry_del_opts(self):
        # Fedora 24 (Workstation Edition)
        osp = get_os_profile_by_id(test_os_id)
        orig_be = create_entry("EDIT_TEST", "2.6.0", "ffffffff",
                               test_lv, lvm_root_lv=test_root_lv,
                               profile=osp)

        # Confirm original entry has been written
        self.assertTrue(exists(orig_be._entry_path))

        # Save these - they will be overwritten by the edit operation
        orig_id = orig_be.boot_id
        orig_entry_path = orig_be._entry_path

        edit_title = "EDITED_TITLE"
        edit_del_opts = "rhgb"

        # FIXME: restore allow_no_dev
        edit_be = edit_entry(Selection(boot_id=orig_id), title=edit_title,
                             del_opts=edit_del_opts)

        # Confirm edited entry has been written
        self.assertTrue(exists(edit_be._entry_path))

        # Confirm original entry has been removed
        self.assertFalse(exists(orig_entry_path))

        # Verify new boot_id
        self.assertFalse(orig_id == edit_be.boot_id)

        # Verify edited title and options
        self.assertEqual(edit_title, edit_be.title)
        self.assertEqual(edit_be.bp.del_opts, [edit_del_opts])
        self.assertTrue(edit_del_opts not in edit_be.options)

        # Clean up entries
        edit_be.delete_entry()

    @unittest.skipIf(not have_root_lv(), "requires root LV")
    def test_edit_entry_del_opts_with_del_opts(self):
        edit_title = "EDITED_TITLE"
        edit_del_opts = "rhgb"
        orig_del_opts = "quiet"

        # Fedora 24 (Workstation Edition)
        osp = get_os_profile_by_id(test_os_id)
        orig_be = create_entry("EDIT_TEST", "2.6.0", "ffffffff",
                               test_lv, lvm_root_lv=test_root_lv,
                               del_opts="quiet", profile=osp)

        # Confirm original entry has been written
        self.assertTrue(exists(orig_be._entry_path))

        # Save these - they will be overwritten by the edit operation
        orig_id = orig_be.boot_id
        orig_entry_path = orig_be._entry_path

        # Verify original deled opts
        self.assertTrue(orig_del_opts not in orig_be.options)
        self.assertEqual(orig_be.bp.del_opts, [orig_del_opts])

        # FIXME: restore allow_no_dev
        edit_be = edit_entry(Selection(boot_id=orig_id), title=edit_title,
                             del_opts=edit_del_opts)

        # Confirm edited entry has been written
        self.assertTrue(exists(edit_be._entry_path))

        # Confirm original entry has been removed
        self.assertFalse(exists(orig_entry_path))

        # Verify new boot_id
        self.assertFalse(orig_id == edit_be.boot_id)

        # Verify edited title and options
        self.assertEqual(edit_title, edit_be.title)

        # Sort the opts lists as Python3 does not guarantee ordering
        sorted_bp_del_opts = sorted(edit_be.bp.del_opts)
        sorted_edit_and_orig_opts = sorted([edit_del_opts, orig_del_opts])
        self.assertEqual(sorted_bp_del_opts, sorted_edit_and_orig_opts)

        # Verify original deleted opts
        self.assertTrue(orig_del_opts not in edit_be.options)
        # Verify edit deleted opts
        self.assertTrue(edit_del_opts not in edit_be.options)

        # Clean up entries
        edit_be.delete_entry()

    @unittest.skipIf(not have_root_lv(), "requires root LV")
    def test_edit_entry_del_opts(self):
        # Fedora 24 (Workstation Edition)
        osp = get_os_profile_by_id(test_os_id)
        orig_be = create_entry("EDIT_TEST", "2.6.0", "ffffffff",
                               test_lv, lvm_root_lv=test_root_lv,
                               profile=osp)

        be = edit_entry(Selection(boot_id=orig_be.boot_id),
                        title="NEWNEWTITLE", del_opts="rhgb quiet")

        self.assertTrue(exists(be._entry_path))
        be.delete_entry()
        self.assertFalse(exists(be._entry_path))

    @unittest.skipIf(not have_root_lv(), "requires root LV")
    def test_edit_delete_entry(self):
        # Fedora 24 (Workstation Edition)
        osp = get_os_profile_by_id(test_os_id)
        orig_be = create_entry("ATITLE", "2.6.0", "ffffffff",
                               test_lv, lvm_root_lv=test_root_lv,
                               profile=osp)
        orig_path = orig_be._entry_path
        self.assertTrue(exists(orig_path))

        edit_be = edit_entry(Selection(boot_id=orig_be.boot_id),
                             title="ANEWTITLE", version="2.6.1")

        self.assertTrue(exists(edit_be._entry_path))
        self.assertFalse(exists(orig_path))

        edit_be.delete_entry()

        self.assertFalse(exists(edit_be._entry_path))

    @unittest.skipIf(not have_root_lv(), "requires root LV")
    def test_edit_entry_no_args(self):
        # Fedora 24 (Workstation Edition)
        osp = get_os_profile_by_id(test_os_id)
        be = create_entry("ATITLE", "2.6.0", "ffffffff", test_lv,
                          lvm_root_lv=test_root_lv, profile=osp)
        self.assertTrue(exists(be._entry_path))

        with self.assertRaises(ValueError) as cm:
            be2 = edit_entry(Selection(boot_id=be.boot_id))

        be.delete_entry()

    @unittest.skipIf(not have_root_lv(), "requires root LV")
    def test_edit_entry_with_add_del_opts(self):
        # Fedora 24 (Workstation Edition)
        osp = get_os_profile_by_id(test_os_id)
        orig_be = create_entry("EDIT_TEST", "2.6.0", "ffffffff",
                               test_lv, lvm_root_lv=test_root_lv,
                               profile=osp)
        orig_path = orig_be._entry_path

        add_opts = "debug"
        del_opts = "rhgb quiet"

        # Entry with options +"debug" -"rhgb quiet"
        orig_boot_id = orig_be.boot_id
        edit_be = edit_entry(Selection(boot_id=orig_boot_id),
                             title="edit with addopts", add_opts=add_opts,
                             del_opts=del_opts)

        self.assertTrue(edit_be)

        self.assertTrue(exists(edit_be._entry_path))
        self.assertFalse(exists(orig_path))

        self.assertTrue(add_opts in edit_be.options)
        self.assertTrue(del_opts not in edit_be.options)

        edit_be.delete_entry()

    def test_print_entries_no_matching(self):
        xoutput = r"BootID.*Version.*Name.*RootDevice"
        output = StringIO()
        opts = BoomReportOpts(report_file=output)
        print_entries(selection=Selection(boot_id="thereisnoboot"), opts=opts)
        self.assertTrue(re.match(xoutput, output.getvalue()))

    def test_print_entries_default_stdout(self):
        print_entries()

    def test_print_entries_boot_id_filter(self):
        xoutput = [r"BootID.*Version.*Name.*RootDevice",
                   r"debfd7f.*4.11.12-100.fc24.x86_64.*Fedora.*"
                   r"/dev/vg00/lvol0-snapshot"]
        output = StringIO()
        opts = BoomReportOpts(report_file=output)
        print_entries(selection=Selection(boot_id="debfd7f"), opts=opts)
        for pair in zip(xoutput, output.getvalue().splitlines()):
            self.assertTrue(re.match(pair[0], pair[1]))

    #
    # API call tests
    #
    # OsProfile tests
    #

    def test_command_create_delete_profile(self):
        osp = create_profile("Some Distro", "somedist", "1 (Qunk)", "1",
                             uname_pattern="sd1",
                             kernel_pattern="/vmlinuz-%{version}",
                             initramfs_pattern="/initramfs-%{version}.img",
                             root_opts_lvm2="rd.lvm.lv=%{lvm_root_lv}",
                             root_opts_btrfs="rootflags=%{btrfs_subvolume}",
                             options="root=%{root_device} %{root_opts}")
        self.assertTrue(osp)
        self.assertEqual(osp.os_name, "Some Distro")

        # Use the OsProfile.delete_profile() method
        osp.delete_profile()

    def test_command_create_delete_profiles(self):
        osp = create_profile("Some Distro", "somedist", "1 (Qunk)", "1",
                             uname_pattern="sd1",
                             kernel_pattern="/vmlinuz-%{version}",
                             initramfs_pattern="/initramfs-%{version}.img",
                             root_opts_lvm2="rd.lvm.lv=%{lvm_root_lv}",
                             root_opts_btrfs="rootflags=%{btrfs_subvolume}",
                             options="root=%{root_device} %{root_opts}")

        self.assertTrue(osp)
        self.assertEqual(osp.os_name, "Some Distro")

        # Use the command.delete_profiles() API call
        delete_profiles(selection=Selection(os_id=osp.os_id))

    def test_command_delete_profiles_no_match(self):
        with self.assertRaises(IndexError) as cm:
            delete_profiles(selection=Selection(os_id="XyZZy"))

    def test_command_create_delete_profile_from_file(self):
        os_release_path = "tests/os-release/fedora26-test-os-release"
        osp = create_profile(None, None, None, None,
                             profile_file=os_release_path, uname_pattern="sd1",
                             kernel_pattern="/vmlinuz-%{version}",
                             initramfs_pattern="/initramfs-%{version}.img",
                             root_opts_lvm2="rd.lvm.lv=%{lvm_root_lv}",
                             root_opts_btrfs="rootflags=%{btrfs_subvolume}",
                             options="root=%{root_device} %{root_opts}")
        self.assertTrue(osp)
        self.assertEqual(osp.os_name, "Fedora")
        self.assertEqual(osp.os_version, "26 (Testing Edition)")
        osp.delete_profile()

    def test_command_create_delete_profile_from_data(self):
        profile_data = {
            BOOM_OS_NAME: "Some Distro", BOOM_OS_SHORT_NAME: "somedist",
            BOOM_OS_VERSION: "1 (Qunk)", BOOM_OS_VERSION_ID: "1",
            BOOM_OS_UNAME_PATTERN: "sd1",
            BOOM_OS_KERNEL_PATTERN: "/vmlinuz-%{version}",
            BOOM_OS_INITRAMFS_PATTERN: "/initramfs-%{version}.img",
            BOOM_OS_ROOT_OPTS_LVM2: "rd.lvm.lv=%{lvm_root_lv}",
            BOOM_OS_ROOT_OPTS_BTRFS: "rootflags=%{btrfs_subvolume}",
            BOOM_OS_OPTIONS: "root=%{root_device} %{root_opts}",
            BOOM_OS_TITLE: "This is a title (%{version})"
        }

        # All fields: success
        osp = create_profile(None, None, None, None, profile_data=profile_data)
        self.assertTrue(osp)
        self.assertEqual(osp.os_name, "Some Distro")
        self.assertEqual(osp.os_version, "1 (Qunk)")
        osp.delete_profile()

        # Pop identity fields in reverse checking order:
        # OS_VERSION_ID, OS_VERSION, OS_SHORT_NAME, OS_NAME

        profile_data.pop(BOOM_OS_VERSION_ID)
        with self.assertRaises(ValueError) as cm:
            bad_osp = create_profile(None, None, None, None,
                                     profile_data=profile_data)

        profile_data.pop(BOOM_OS_VERSION)
        with self.assertRaises(ValueError) as cm:
            bad_osp = create_profile(None, None, None, None,
                                     profile_data=profile_data)

        profile_data.pop(BOOM_OS_SHORT_NAME)
        with self.assertRaises(ValueError) as cm:
            bad_osp = create_profile(None, None, None, None,
                                     profile_data=profile_data)

        profile_data.pop(BOOM_OS_NAME)
        with self.assertRaises(ValueError) as cm:
            bad_osp = create_profile(None, None, None, None,
                                     profile_data=profile_data)

    def test_clone_profile_no_os_id(self):
        with self.assertRaises(ValueError) as cm:
            bad_osp = clone_profile(Selection())

    def test_clone_profile_no_args(self):
        with self.assertRaises(ValueError) as cm:
            bad_osp = clone_profile(Selection(os_id="d4439b7"))

    def test_clone_profile_no_matching_os_id(self):
        with self.assertRaises(ValueError) as cm:
            bad_osp = clone_profile(Selection(os_id="fffffff"), name="NEW")

    def test_clone_profile_ambiguous_os_id(self):
        with self.assertRaises(ValueError) as cm:
            bad_osp = clone_profile(Selection(os_id="d"), name="NEW")

    def test_clone_profile_new_name(self):
        osp = clone_profile(Selection(os_id="d4439b7"),
                            name="NEW", short_name="new", version="26 (Not)",
                            version_id="~26")
        self.assertTrue(osp)
        self.assertEqual("NEW", osp.os_name)
        self.assertEqual("new", osp.os_short_name)
        osp.delete_profile()

    def test_create_edit_profile(self):
        osp = create_profile("Test1", "test", "1 (Test)", "1",
                             uname_pattern="t1")

        self.assertTrue(osp)

        edit_osp = edit_profile(Selection(os_id=osp.os_id),
                                uname_pattern="t2")

        self.assertTrue(edit_osp)
        self.assertEqual(osp.uname_pattern, "t2")
        osp.delete_profile()
        edit_osp.delete_profile()

    def test_edit_no_matching_os_id(self):
        with self.assertRaises(ValueError) as cm:
            edit_osp = edit_profile(Selection(os_id="notfound"),
                                    uname_pattern="nf2")

    def test_edit_ambiguous_os_id(self):
        with self.assertRaises(ValueError) as cm:
            edit_osp = edit_profile(Selection(os_id="d"),
                                    uname_pattern="d2")

    def test_list_profiles(self):
        profiles = list_profiles()
        self.assertTrue(profiles)

    def test_print_profiles(self):
        repstr = print_profiles()

    #
    # API call tests
    #
    # HostProfile tests
    #

    def test_create_delete_host(self):
        osp = create_profile("Some Distro", "somedist", "1 (Qunk)", "1",
                             uname_pattern="sd1",
                             kernel_pattern="/vmlinuz-%{version}",
                             initramfs_pattern="/initramfs-%{version}.img",
                             root_opts_lvm2="rd.lvm.lv=%{lvm_root_lv}",
                             root_opts_btrfs="rootflags=%{btrfs_subvolume}",
                             options="root=%{root_device} %{root_opts}")

        self.assertTrue(osp)
        self.assertEqual(osp.os_name, "Some Distro")

        host_machine_id = "ffffffffffffffff1234567890"
        host_name = "somehost.somedomain"
        host_opts = osp.options + " hostoptions"

        hp = create_host(machine_id=host_machine_id, host_name=host_name,
                         os_id=osp.os_id, label="", options=host_opts)

        self.assertEqual(host_machine_id, hp.machine_id)
        self.assertEqual(host_name, hp.host_name)
        self.assertEqual(host_opts, hp.options)

        # Use the command.delete_hosts() API call
        delete_hosts(Selection(host_id=hp.host_id))

        # Clean up osp
        osp.delete_profile()

    def test_create_host_no_os_id(self):
        os_id = None
        host_machine_id = "ffffffffffffffff1234567890"
        host_name = "somehost.somedomain"
        host_opts = "hostoptions"

        with self.assertRaises(ValueError) as cm:
            bad_hp = create_host(machine_id=host_machine_id,
                                 host_name=host_name, os_id=os_id,
                                 label="", options=host_opts)

    def test_create_host_no_os_id_match(self):
        os_id = "notfound"
        host_machine_id = "ffffffffffffffff1234567890"
        host_name = "somehost.somedomain"
        host_opts = "hostoptions"

        with self.assertRaises(ValueError) as cm:
            bad_hp = create_host(machine_id=host_machine_id,
                                 host_name=host_name, os_id=os_id,
                                 label="", options=host_opts)

    def test_create_host_no_host_name(self):
        osp = create_profile("Some Distro", "somedist", "1 (Qunk)", "1",
                             uname_pattern="sd1",
                             kernel_pattern="/vmlinuz-%{version}",
                             initramfs_pattern="/initramfs-%{version}.img",
                             root_opts_lvm2="rd.lvm.lv=%{lvm_root_lv}",
                             root_opts_btrfs="rootflags=%{btrfs_subvolume}",
                             options="root=%{root_device} %{root_opts}")

        self.assertTrue(osp)
        self.assertEqual(osp.os_name, "Some Distro")

        host_machine_id = "ffffffffffffffff1234567890"
        host_name = ""
        host_opts = "hostoptions"

        with self.assertRaises(ValueError) as cm:
            bad_hp = create_host(machine_id=host_machine_id,
                                 host_name=host_name, os_id=osp.os_id,
                                 label="", options=host_opts)

        osp.delete_profile()

    def test_create_host_no_machine_id(self):
        osp = create_profile("Some Distro", "somedist", "1 (Qunk)", "1",
                             uname_pattern="sd1",
                             kernel_pattern="/vmlinuz-%{version}",
                             initramfs_pattern="/initramfs-%{version}.img",
                             root_opts_lvm2="rd.lvm.lv=%{lvm_root_lv}",
                             root_opts_btrfs="rootflags=%{btrfs_subvolume}",
                             options="root=%{root_device} %{root_opts}")

        self.assertTrue(osp)
        self.assertEqual(osp.os_name, "Some Distro")

        host_machine_id = ""
        host_name = "somehost.somedomain"
        host_opts = "hostoptions"

        with self.assertRaises(ValueError) as cm:
            bad_hp = create_host(machine_id=host_machine_id,
                                 host_name=host_name, os_id=osp.os_id,
                                 label="", options=host_opts)

        osp.delete_profile()

    def test_create_host_all_args(self):
        osp = create_profile("Some Distro", "somedist", "1 (Qunk)", "1",
                             uname_pattern="sd1",
                             kernel_pattern="/vmlinuz-%{version}",
                             initramfs_pattern="/initramfs-%{version}.img",
                             root_opts_lvm2="rd.lvm.lv=%{lvm_root_lv}",
                             root_opts_btrfs="rootflags=%{btrfs_subvolume}",
                             options="root=%{root_device} %{root_opts}")

        self.assertTrue(osp)
        self.assertEqual(osp.os_name, "Some Distro")

        host_machine_id = "ffffffffffffffff1234567890"
        host_name = "somehost.somedomain"

        hp = create_host(machine_id=host_machine_id, host_name=host_name,
                         os_id=osp.os_id, label="label",
                         kernel_pattern="/vmlinuz",
                         initramfs_pattern="/initramfs.img",
                         root_opts_lvm2="rd.lvm.lv=vg/lv",
                         root_opts_btrfs="rootflags=subvolid=1",
                         options=osp.options, add_opts="debug",
                         del_opts="rhgb quiet")

        self.assertEqual(host_machine_id, hp.machine_id)
        self.assertEqual(host_name, hp.host_name)

        hp.delete_profile()

        # Clean up osp
        osp.delete_profile()

    def test_delete_hosts_no_match(self):
        with self.assertRaises(IndexError) as cm:
            delete_hosts(Selection(host_id="nomatch"))

    def test_clone_host(self):
        osp = create_profile("Some Distro", "somedist", "1 (Qunk)", "1",
                             uname_pattern="sd1",
                             kernel_pattern="/vmlinuz-%{version}",
                             initramfs_pattern="/initramfs-%{version}.img",
                             root_opts_lvm2="rd.lvm.lv=%{lvm_root_lv}",
                             root_opts_btrfs="rootflags=%{btrfs_subvolume}",
                             options="root=%{root_device} %{root_opts}")

        self.assertTrue(osp)
        self.assertEqual(osp.os_name, "Some Distro")

        host_machine_id = "ffffffffffffffff1234567890"
        clone_machine_id = "ffffffffffffffff0987654321"
        host_name = "somehost.somedomain"
        host_opts = osp.options + " hostoptions"

        hp = create_host(machine_id=host_machine_id, host_name=host_name,
                         os_id=osp.os_id, label="", options=host_opts)

        self.assertEqual(host_machine_id, hp.machine_id)
        self.assertEqual(host_name, hp.host_name)
        self.assertEqual(host_opts, hp.options)

        clone_hp = clone_host(Selection(host_id=hp.host_id),
                              machine_id=clone_machine_id)

        self.assertEqual(clone_machine_id, clone_hp.machine_id)
        self.assertNotEqual(hp.host_id, clone_hp.host_id)

        hp.delete_profile()
        clone_hp.delete_profile()

        # Clean up osp
        osp.delete_profile()

    def test_clone_host_no_host_id(self):
        with self.assertRaises(ValueError) as cm:
            bad_hp = clone_host(Selection(host_id=None))

    def test_clone_host_no_host_id_match(self):
        host_id = "notfound"

        with self.assertRaises(ValueError) as cm:
            bad_hp = clone_host(Selection(host_id=host_id),
                                machine_id="ffffffff")

    def test_clone_host_no_args(self):
        host_id = "5ebcb1f"

        with self.assertRaises(ValueError) as cm:
            bad_hp = clone_host(Selection(host_id=host_id))

    def test_create_edit_host(self):
        osp = create_profile("Some Distro", "somedist", "1 (Qunk)", "1",
                             uname_pattern="sd1",
                             kernel_pattern="/vmlinuz-%{version}",
                             initramfs_pattern="/initramfs-%{version}.img",
                             root_opts_lvm2="rd.lvm.lv=%{lvm_root_lv}",
                             root_opts_btrfs="rootflags=%{btrfs_subvolume}",
                             options="root=%{root_device} %{root_opts}")

        self.assertTrue(osp)
        self.assertEqual(osp.os_name, "Some Distro")

        host_machine_id = "ffffffffffffffff1234567890"
        host_name = "somehost.somedomain"
        host_opts = osp.options + " hostoptions"

        hp = create_host(machine_id=host_machine_id, host_name=host_name,
                         os_id=osp.os_id, label="", options=host_opts)

        self.assertEqual(host_machine_id, hp.machine_id)
        self.assertEqual(host_name, hp.host_name)
        self.assertEqual(host_opts, hp.options)

        edit_name = "someother.host"
        edit_opts = osp.options

        edit_hp = edit_host(Selection(host_id=hp.host_id),
                            machine_id=host_machine_id, host_name=edit_name,
                            os_id=osp.os_id, label="", options=edit_opts)

        self.assertEqual(host_machine_id, edit_hp.machine_id)
        self.assertEqual(edit_name, edit_hp.host_name)
        self.assertEqual(osp.options, edit_hp.options)

        edit_hp.delete_profile()

        # Clean up osp
        osp.delete_profile()

    def test_list_hosts_default(self):
        """Test the list_hosts() API call with no selection.
        """
        hps = list_hosts()
        self.assertTrue(len(hps) >= 1)

    def test_print_hosts_default(self):
        """Test the list_hosts() API call with no selection.
        """
        print_hosts()

    #
    # Command handler tests
    #

    def test__create_cmd(self):
        """Test the _create_cmd() handler with correct arguments.
        """
        args = get_create_cmd_args()
        opts = boom.command._report_opts_from_args(args)
        boom.command._create_cmd(args, None, opts, None)

    def test__create_cmd_bad_identity(self):
        """Test the _create_cmd() handler with an invalid identity
            function argument.
        """
        args = get_create_cmd_args()
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._create_cmd(args, None, opts, "badident")
        self.assertEqual(r, 1)

    @unittest.skip("Requires boom.command.get_uts_release() override")
    def test__create_cmd_no_version(self):
        """Test the _create_cmd() handler with missing version.
        """
        args = get_create_cmd_args()
        args.version = None
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._create_cmd(args, None, opts, None)
        self.assertEqual(r, 1)

    @unittest.skipIf(not have_root_lv(), "requires root LV")
    def test__create_cmd_version_from_uts(self):
        """Test the _create_cmd() handler with missing version, and the
            default version obtained from the system UTS data.
        """
        args = get_create_cmd_args()
        args.version = None
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._create_cmd(args, None, opts, None)
        self.assertNotEqual(r, 1)

    def test__create_cmd_no_root_device(self):
        """Test the _create_cmd() handler with missing root device.
        """
        args = get_create_cmd_args()
        args.root_device = None
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._create_cmd(args, None, opts, None)
        self.assertEqual(r, 1)

    @unittest.skipIf(not have_root_lv(), "requires root LV")
    def test__create_cmd_auto_machine_id(self):
        """Test the _create_cmd() handler with automatic machine_id.
        """
        args = get_create_cmd_args()
        args.machine_id = None
        args.profile = None
        args.version = "5.8.16-200.fc32.x86_64"
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._create_cmd(args, None, opts, None)
        self.assertNotEqual(r, 1)

    def test__create_cmd_no_profile(self):
        """Test the _create_cmd() handler with missing profile.
        """
        args = get_create_cmd_args()
        args.profile = None
        # Avoid HostProfile match
        args.machine_id = "quxquxquxqux"
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._create_cmd(args, None, opts, None)
        self.assertEqual(r, 1)

    def test__create_cmd_no_title(self):
        """Test the _create_cmd() handler with missing title.
        """
        args = get_create_cmd_args()
        args.title = None

        # Avoid OsProfile auto-title
        osp = get_os_profile_by_id(test_os_id)
        osp.title = None

        opts = boom.command._report_opts_from_args(args)
        r = boom.command._create_cmd(args, None, opts, None)
        self.assertEqual(r, 1)

    def test__delete_cmd_no_selection(self):
        """Test that _delete_cmd() rejects a call with no valid
            selection.
        """
        args = MockArgs()
        args.boot_id = None
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._delete_cmd(args, None, opts, None)
        self.assertEqual(r, 1)

    @unittest.skipIf(not have_root_lv(), "requires root LV")
    def test__create_cmd_with_override(self):
        args = get_create_cmd_args()
        args.title = "override test"
        # Use a profile that includes BOOT_IMAGE=%{kernel} in BOOM_OS_OPTIONS
        args.profile = "d4439b7"
        # Use an image string ("vmlinux") that does not match the OsProfile
        # template pattern for a Linux bzImage ("vmlinu*z*").
        args.linux = "/vmzlinux-test"
        args.initrd = "/initrd-test.img"
        opts = boom.command._report_opts_from_args(args)
        boom.command._create_cmd(args, None, opts, None)

        # Find entry and verify --linux and --initrd override
        be = find_entries(Selection(title=args.title))[0]
        boot_id = be.boot_id
        self.assertEqual(be.linux, args.linux)
        self.assertEqual(be.initrd, args.initrd)

        # Reload entry and verify boot_id and overrides
        drop_entries()
        load_entries()
        self.assertEqual(be.boot_id, boot_id)
        self.assertEqual(be.linux, args.linux)
        self.assertEqual(be.initrd, args.initrd)

    def test__delete_cmd(self):
        """Test the _delete_cmd() handler with a valid entry.
        """
        args = MockArgs()
        args.boot_id = "61bcc49"
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._delete_cmd(args, None, opts, None)
        self.assertNotEqual(r, 1)

    def test__delete_cmd_no_selection(self):
        """Test the _delete_cmd() handler with no valid entry selection.
        """
        args = MockArgs()
        args.boot_id = None
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._delete_cmd(args, None, opts, None)
        self.assertEqual(r, 1)

    def test__delete_cmd_verbose(self):
        """Test the _delete_cmd() handler with a valid entry.
        """
        args = MockArgs()
        args.boot_id = "61bcc49"
        args.verbose = 1 # enable reporting
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._delete_cmd(args, None, opts, None)
        self.assertNotEqual(r, 1)

    def test__delete_cmd_with_options(self):
        """Test the _delete_cmd() handler with a valid entry and report
            options object setting columns-as-rows mode.
        """
        args = MockArgs()
        args.boot_id = "61bcc49"
        opts = boom.command._report_opts_from_args(args)
        opts.columns_as_rows = True
        r = boom.command._delete_cmd(args, None, opts, None)
        self.assertEqual(r, 0)

    def test__delete_cmd_with_fields(self):
        """Test the _delete_cmd() handler with a valid entry and report
            field options string.
        """
        args = MockArgs()
        args.boot_id = "61bcc49"
        args.options = "title,bootid"
        args.verbose = 1 # enable reporting
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._delete_cmd(args, None, opts, None)
        self.assertEqual(r, 0)

    def test__delete_cmd_with_bad_fields(self):
        """Test the _delete_cmd() handler with a valid entry and invalid
            report field options string.
        """
        args = MockArgs()
        args.boot_id = "61bcc49"
        opts = boom.command._report_opts_from_args(args)
        args.options = "I,wish,I,knew,how,it,would,feel,to,be,free"
        args.verbose = 1 # enable reporting
        r = boom.command._delete_cmd(args, None, opts, None)
        self.assertEqual(r, 1)

    def test__delete_cmd_verbose(self):
        """Test the _delete_cmd() handler with a valid entry and
            verbose output.
        """
        args = MockArgs()
        args.boot_id = "61bcc49"
        args.verbose = 1
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._delete_cmd(args, None, opts, None)
        self.assertNotEqual(r, 1)

    def test__delete_cmd_identity(self):
        """Test the _delete_cmd() handler with a valid entry that
            is passed via the 'identiry' handler argument.
        """
        args = MockArgs()
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._delete_cmd(args, None, opts, "61bcc49")
        self.assertNotEqual(r, 1)

    def test__delete_cmd_no_criteria(self):
        """Test the _delete_cmd() handler with no valid selection.
        """
        args = MockArgs()
        args.boot_id = None
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._delete_cmd(args, None, opts, None)
        self.assertEqual(r, 1)

    def test__delete_cmd_multi(self):
        """Test the _delete_cmd() handler with multiple valid entries.
        """
        args = MockArgs()
        args.boot_id = "6" # Matches four entries
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._delete_cmd(args, None, opts, None)
        self.assertNotEqual(r, 1)

    def test__delete_cmd_no_matching(self):
        """Test the _delete_cmd() handler with no matching entries.
        """
        args = MockArgs()
        args.boot_id = "qux" # Matches no entries
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._delete_cmd(args, None, opts, None)
        self.assertEqual(r, 1)

    def test__clone_cmd(self):
        """Test the _clone_cmd() handler with a valid entry and new
            title.
        """
        args = MockArgs()
        args.boot_id = "61bcc49"
        args.title = "Something New"
        # Disable device presence checks
        args.no_dev = True
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._clone_cmd(args, None, opts, None)
        self.assertNotEqual(r, 1)

    def test__clone_cmd_no_criteria(self):
        """Test the _clone_cmd() handler with no valid selection.
        """
        args = MockArgs()
        args.boot_id = None
        args.title = "Something New"
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._clone_cmd(args, None, opts, None)
        self.assertEqual(r, 1)

    def test__clone_cmd_no_matching(self):
        """Test the _clone_cmd() handler with no matching entries.
        """
        args = MockArgs()
        args.boot_id = "qux"
        args.title = "Something New"
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._clone_cmd(args, None, opts, None)
        self.assertEqual(r, 1)

    def test__show_cmd(self):
        """Test the _show_cmd() handler.
        """
        args = MockArgs()
        r = boom.command._show_cmd(args, None, None, None)
        self.assertEqual(r, 0)

    def test__show_cmd_single(self):
        """Test the _show_cmd() handler with a single selected entry.
        """
        args = MockArgs()
        args.boot_id = "61bcc49"
        r = boom.command._show_cmd(args, None, None, None)
        self.assertEqual(r, 0)

    def test__show_cmd_single_identifier(self):
        """Test the _show_cmd() handler with a single identifier.
        """
        args = MockArgs()
        r = boom.command._show_cmd(args, None, None, "61bcc49")
        self.assertEqual(r, 0)

    def test__show_cmd_selection(self):
        """Test the _show_cmd() handler with multiple selected entries.
        """
        args = MockArgs()
        args.boot_id = "6" # Matches four entries
        r = boom.command._show_cmd(args, None, None, None)
        self.assertEqual(r, 0)

    def test__show_cmd_invalid_selection(self):
        """Test the _show_cmd() handler with an invalid selection.
        """
        args = MockArgs()
        # Clear boot_id
        args.boot_id = None
        # Invalid selection criteria for BootEntry type
        select = Selection(host_add_opts="qux")
        r = boom.command._show_cmd(args, select, None, None)
        self.assertEqual(r, 1)

    def test__list_cmd(self):
        args = MockArgs()
        r = boom.command._list_cmd(args, None, None, None)
        self.assertNotEqual(r, 1)

    def test__list_cmd_single(self):
        args = MockArgs()
        args.boot_id = "61bcc49"
        r = boom.command._list_cmd(args, None, None, None)
        self.assertNotEqual(r, 1)

    def test__list_cmd_single_identifier(self):
        """Test the _list_cmd() handler with a single identifier.
        """
        args = MockArgs()
        r = boom.command._list_cmd(args, None, None, "61bcc49")
        self.assertEqual(r, 0)

    def test__list_cmd_selection(self):
        """Test the _list_cmd() handler with multiple selected entries.
        """
        args = MockArgs()
        args.boot_id = "6" # Matches four entries
        r = boom.command._list_cmd(args, None, None, None)
        self.assertEqual(r, 0)

    def test__list_cmd_invalid_selection(self):
        """Test the _list_cmd() handler with an invalid selection.
        """
        args = MockArgs()
        # Clear boot_id
        args.boot_id = None
        # Invalid selection criteria for BootEntry type
        select = Selection(host_add_opts="qux")
        r = boom.command._list_cmd(args, select, None, None)
        self.assertEqual(r, 1)

    def test__list_cmd_with_options(self):
        """Test the _list_cmd() handler with report field options
            string.
        """
        args = MockArgs()
        args.options = "title"
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._list_cmd(args, None, opts, None)
        self.assertEqual(r, 0)

    def test__list_cmd_with_sort(self):
        args = MockArgs()
        args.sort = "bootid"
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._list_cmd(args, Selection(), opts, None)
        self.assertEqual(r, 0)

    def test__list_cmd_with_sort_ascending(self):
        args = MockArgs()
        args.sort = "+bootid"
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._list_cmd(args, Selection(), opts, None)
        self.assertEqual(r, 0)

    def test__list_cmd_with_sort_descending(self):
        args = MockArgs()
        args.sort = "-bootid"
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._list_cmd(args, Selection(), opts, None)
        self.assertEqual(r, 0)

    def test__list_cmd_with_sort_two(self):
        args = MockArgs()
        args.sort = "bootid,version"
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._list_cmd(args, Selection(), opts, None)
        self.assertEqual(r, 0)

    def test__list_cmd_with_sort_bad(self):
        args = MockArgs()
        args.sort = "qux"
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._list_cmd(args, Selection(), opts, None)
        self.assertEqual(r, 1)

    def test__list_cmd_with_options_help(self):
        args = MockArgs()
        args.options = "help"
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._list_cmd(args, Selection(), opts, None)
        self.assertEqual(r, 0)

    def test__list_cmd_verbose(self):
        """Test the _list_cmd() handler with a valid entry and
            verbose output.
        """
        args = MockArgs()
        args.boot_id = "61bcc49"
        args.verbose = 1
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._list_cmd(args, None, opts, None)
        self.assertEqual(r, 0)

    def test__edit_cmd(self):
        """Test the _edit_cmd() handler with a valid entry and new
            title.
        """
        args = MockArgs()
        args.boot_id = "61bcc49"
        args.title = "Something New"
        # Disable device presence checks
        args.no_dev = True
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._edit_cmd(args, None, opts, None)
        self.assertNotEqual(r, 1)

    def test__edit_cmd_no_criteria(self):
        """Test the _edit_cmd() handler with no valid selection.
        """
        args = MockArgs()
        args.boot_id = None
        args.title = "Something New"
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._edit_cmd(args, None, opts, None)
        self.assertEqual(r, 1)

    def test__edit_cmd_no_matching(self):
        """Test the _edit_cmd() handler with no matching entries.
        """
        args = MockArgs()
        args.boot_id = "qux"
        args.title = "Something New"
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._edit_cmd(args, None, opts, None)
        self.assertEqual(r, 1)

    def test__create_profile_cmd_bad_identity(self):
        """Test the _create_profile_cmd() handler with a non-None
            identity argument.
        """
        args = MockArgs()
        r = boom.command._create_profile_cmd(args, None, None, "12345")
        self.assertEqual(r, 1)

    def test__create_profile_cmd(self):
        """Test the _create_profile_cmd() handler with valid args.
        """
        args = MockArgs()
        args.name = "Test OS"
        args.short_name = "testos"
        args.os_version = "1 (Workstation)"
        args.os_version_id = "1"
        args.uname_pattern = "to1"
        r = boom.command._create_profile_cmd(args, None, None, None)
        self.assertEqual(r, 0)

    def test__create_profile_cmd_no_name(self):
        """Test the _create_profile_cmd() handler with valid args.
        """
        args = MockArgs()
        args.name = None
        args.short_name = "testos"
        args.os_version = "1 (Workstation)"
        args.os_version_id = "1"
        args.uname_pattern = "to1"
        r = boom.command._create_profile_cmd(args, None, None, None)
        self.assertEqual(r, 1)

    def test__create_profile_cmd_no_short_name(self):
        """Test the _create_profile_cmd() handler with valid args.
        """
        args = MockArgs()
        args.name = "Test OS"
        args.short_name = None
        args.os_version = "1 (Workstation)"
        args.os_version_id = "1"
        args.uname_pattern = "to1"
        r = boom.command._create_profile_cmd(args, None, None, None)
        self.assertEqual(r, 1)

    def test__create_profile_cmd_no_version(self):
        """Test the _create_profile_cmd() handler with valid args.
        """
        args = MockArgs()
        args.name = "Test OS"
        args.short_name = "testos"
        args.os_version = None
        args.os_version_id = "1"
        args.uname_pattern = "to1"
        r = boom.command._create_profile_cmd(args, None, None, None)
        self.assertEqual(r, 1)

    def test__create_profile_cmd_no_version_id(self):
        """Test the _create_profile_cmd() handler with valid args.
        """
        args = MockArgs()
        args.name = "Test OS"
        args.short_name = "testos"
        args.os_version = "1 (Workstation)"
        args.os_version_id = None
        args.uname_pattern = "to1"
        r = boom.command._create_profile_cmd(args, None, None, None)
        self.assertEqual(r, 1)

    def test__create_profile_cmd_no_uname_pattern(self):
        """Test the _create_profile_cmd() handler with valid args.
        """
        args = MockArgs()
        args.name = "Test OS"
        args.short_name = "testos"
        args.os_version = "1 (Workstation)"
        args.os_version_id = "1"
        args.uname_pattern = None
        r = boom.command._create_profile_cmd(args, None, None, None)
        self.assertEqual(r, 1)

    def test__create_profile_cmd_from_host(self):
        """Test that creation of an OsProfile from /etc/os-release on
            the running host succeeds.
        """
        # Depending on the machine the test suite is running on it is
        # possible that an OsProfile already exists for the system. To
        # avoid a collision between an existing host OsProfile and the
        # newly created test profile, attempt to delete any existing
        # profile from the test sandbox first.
        drop_profiles()
        host_os_id = OsProfile.from_host_os_release().os_id
        load_profiles()
        if host_os_id:
            try:
                delete_profiles(selection=Selection(os_id=host_os_id))
            except Exception:
                pass

        args = MockArgs()
        args.uname_pattern = "test1"
        args.from_host = True
        r = boom.command._create_profile_cmd(args, None, None, None)
        self.assertEqual(r, 0)

    def test__create_profile_cmd_from_os_release(self):
        """Test creation of an OsProfile from an os-release file.
        """
        test_os_release = "tests/os-release/test-os-release"
        args = MockArgs()
        args.uname_pattern = "test1"
        args.os_release = test_os_release
        r = boom.command._create_profile_cmd(args, None, None, None)
        self.assertEqual(r, 0)

    def test__create_profile_cmd_invalid_identifier(self):
        """Test that _create_profile_cmd() rejects an identifier arg.
        """
        args = MockArgs()
        identifier = "d4439b7"
        r = boom.command._create_profile_cmd(args, None, None, identifier)
        self.assertEqual(r, 1)

    def test__delete_profile_cmd_valid_identifier(self):
        """Test that _delete_profile_cmd() deletes a profile via a
            valid identifier arg.
        """
        args = MockArgs()
        identifier = "d4439b7"
        r = boom.command._delete_profile_cmd(args, None, None, identifier)
        self.assertEqual(r, 0)

    def test__delete_profile_cmd_no_selection(self):
        """Test that _delete_profile_cmd() returns an error with no
            profile selection.
        """
        args = MockArgs()
        args.profile = None
        r = boom.command._delete_profile_cmd(args, None, None, None)
        self.assertEqual(r, 1)

    def test__delete_profiles_cmd_verbose(self):
        """Test the _delete_profile_cmd() handler with reporting.
        """
        args = MockArgs()
        args.profile = "d4439b7"
        args.verbose = 1 # enable reporting
        r = boom.command._delete_profile_cmd(args, None, None, None)
        self.assertEqual(r, 0)

    def test__delete_profiles_cmd_with_fields(self):
        """Test the _delete_profile_cmd() handler with reporting.
        """
        args = MockArgs()
        args.profile = "d4439b7"
        args.options = "osid,osname"
        args.verbose = 1 # enable reporting
        r = boom.command._delete_profile_cmd(args, None, None, None)
        self.assertEqual(r, 0)

    def test__delete_profiles_cmd_with_bad_fields(self):
        """Test the _delete_profile_cmd() handler with reporting.
        """
        args = MockArgs()
        args.profile = "d4439b7"
        args.options = "There,is,water,at,the,bottom,of,the,ocean"
        args.verbose = 1 # enable reporting
        r = boom.command._delete_profile_cmd(args, None, None, None)
        self.assertEqual(r, 1)

    def test__clone_profile_cmd(self):
        """Test the _clone_profile_cmd() handler with a valid os_id and
            new name.
        """
        args = MockArgs()
        args.profile = "d4439b7"
        args.short_name = "somethingsomethingsomething profile side"
        r = boom.command._clone_profile_cmd(args, None, None, None)
        self.assertEqual(r, 0)

    def test__clone_profile_cmd_no_criteria(self):
        """Test the _clone_profile_cmd() handler with no valid selection.
        """
        args = MockArgs()
        args.profile = None
        args.name = "Something Something Something, Profile Side"
        r = boom.command._clone_profile_cmd(args, None, None, None)
        self.assertEqual(r, 1)

    def test__clone_profile_cmd_no_matching(self):
        """Test the _clone_profile_cmd() handler with no matching entries.
        """
        args = MockArgs()
        args.profile = "thisisnottheprofileyouarelookingfor"
        args.name = "Something Something Something, Profile Side"
        r = boom.command._clone_profile_cmd(args, None, None, None)
        self.assertEqual(r, 1)

    def test__show_profile_cmd(self):
        """Test the _show_profile() command handler with defaults args.
        """
        args = MockArgs()
        r = boom.command._show_profile_cmd(args, None, None, None)
        self.assertEqual(r, 0)

    def test__show_profile_cmd_with_identifier(self):
        """Test the _show_profile() command handler with defaults args.
        """
        args = MockArgs()
        os_id = "d4439b7"
        r = boom.command._show_profile_cmd(args, None, None, os_id)
        self.assertEqual(r, 0)

    def test__show_profile_cmd_with_profile_arg(self):
        """Test the _show_profile() command handler with defaults args.
        """
        args = MockArgs()
        args.profile = "d4439b7"
        r = boom.command._show_profile_cmd(args, None, None, None)
        self.assertEqual(r, 0)

    def test__list_profile_cmd(self):
        args = MockArgs()
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._list_profile_cmd(args, None, opts, None)
        self.assertEqual(r, 0)

    def test__list_profile_cmd_with_identifier(self):
        args = MockArgs()
        os_id = "d4439b7"
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._list_profile_cmd(args, None, opts, os_id)
        self.assertEqual(r, 0)

    def test__list_profile_cmd_with_profile_arg(self):
        args = MockArgs()
        args.profile = "d4439b7"
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._list_profile_cmd(args, None, opts, None)
        self.assertEqual(r, 0)

    def test__list_profile_cmd_with_options(self):
        """Test the _list_cmd() handler with report field options
            string.
        """
        args = MockArgs()
        args.options = "osname,osversion"
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._list_profile_cmd(args, None, opts, None)
        self.assertEqual(r, 0)

    def test__list_profile_cmd_with_verbose(self):
        """Test the _list_cmd() handler with report field options
            string.
        """
        args = MockArgs()
        args.verbose = 1
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._list_profile_cmd(args, None, opts, None)
        self.assertEqual(r, 0)

    def test__edit_profile_cmd(self):
        """Test the _edit_profile_cmd() hander with default args.
        """
        args = MockArgs()
        args.profile = "d4439b7"
        args.uname_pattern = "nf26"
        args.os_options = "root=%{root_device} boot and stuff"
        r = boom.command._edit_profile_cmd(args, None, None, None)
        self.assertEqual(r, 0)

    def test__edit_profile_cmd_with_identifier(self):
        """Test the _edit_profile_cmd() handler with an identifier.
        """
        args = MockArgs()
        os_id = "d4439b7"
        args.uname_pattern = "nf26"
        args.os_options = "root=%{root_device} boot and stuff"
        r = boom.command._edit_profile_cmd(args, None, None, os_id)
        self.assertEqual(r, 0)

    def test__edit_profile_cmd_ambiguous_identifier(self):
        """Test the _edit_profile_cmd() handler with an ambiguous
            identifier argument.
        """
        args = MockArgs()
        os_id = "d"
        args.uname_pattern = "nf26"
        args.os_options = "boot and stuff"
        r = boom.command._edit_profile_cmd(args, None, None, os_id)
        self.assertEqual(r, 1)

    def test__edit_profile_cmd_with_options(self):
        """Test the _edit_profile_cmd() handler with report control
            options.
        """
        args = MockArgs()
        args.profile = "d4439b7"
        args.options = "badoptions"
        r = boom.command._edit_profile_cmd(args, None, None, None)
        self.assertEqual(r, 1)

    def test__edit_profile_cmd_edits_identity_keys(self):
        """Test the _edit_profile_cmd() handler with invalid profile
            key modifications.
        """
        args = MockArgs()
        args.profile = "d4439b7"
        # Can only change via clone
        args.name = "Bad Fedora"
        r = boom.command._edit_profile_cmd(args, None, None, None)
        self.assertEqual(r, 1)

    def test__clone_profile_cmd(self):
        """Test the _clone_profile_cmd() handler with valid args.
        """
        args = MockArgs()
        args.profile = "d4439b7"
        args.name = "NotFedora"
        args.short_name = "notfedora"
        r = boom.command._clone_profile_cmd(args, None, None, None)
        self.assertEqual(r, 0)

    def test__create_host_cmd_with_identifier(self):
        """Test _create_host_cmd() with an invalid identifier arg.
        """
        args = MockArgs()
        identifier = "badidentity"
        r = boom.command._create_host_cmd(args, None, None, identifier)
        self.assertEqual(r, 1)

    def test__create_host_cmd_no_name(self):
        """Test the _create_host_cmd() handler with no name argument.
        """
        args = MockArgs()
        args.host_name = None
        r = boom.command._create_host_cmd(args, None, None, None)
        self.assertEqual(r, 1)

    def test__create_host_cmd_no_profile(self):
        """Test the _clone_profile_cmd() handler with missing profile
            argument.
        """
        args = MockArgs()
        args.name = "NotFedora"
        args.short_name = "notfedora"
        r = boom.command._create_host_cmd(args, None, None, None)
        self.assertEqual(r, 1)

    def test__create_host_cmd(self):
        """Test the _create_host_cmd() handler with valid args.
        """
        args = MockArgs()
        args.machine_id = "611f38fd887d41fffffffffffffff000"
        args.host_name = "newhost"
        args.profile = "d4439b7"
        r = boom.command._create_host_cmd(args, None, None, None)
        self.assertEqual(r, 0)

    def test__delete_host_cmd(self):
        """Test the _delete_host_cmd() handler with valid --host-id
            argument.
        """
        args = MockArgs()
        args.host_id = "5ebcb1f"
        r = boom.command._delete_host_cmd(args, None, None, None)
        self.assertEqual(r, 0)

    def test__delete_host_cmd_with_options(self):
        """Test the _delete_host_cmd() handler with valid --host-id
            argument and report control options.
        """
        args = MockArgs()
        args.host_id = "5ebcb1f"
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._delete_host_cmd(args, None, opts, None)
        self.assertEqual(r, 0)

    def test__delete_host_cmd_with_verbose(self):
        """Test the _delete_host_cmd() handler with valid --host-id
            argument and verbosity.
        """
        args = MockArgs()
        args.host_id = "5ebcb1f"
        args.verbose = 1
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._delete_host_cmd(args, None, opts, None)
        self.assertEqual(r, 0)

    def test__delete_host_cmd_with_fields(self):
        """Test the _delete_host_cmd() handler with valid --host-id
            argument and custom report field options.
        """
        args = MockArgs()
        args.host_id = "5ebcb1f"
        args.options = "hostid,hostname"
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._delete_host_cmd(args, None, opts, None)
        self.assertEqual(r, 0)

    def test__delete_host_cmd_with_identifier(self):
        """Test the _delete_host_cmd() handler with valid identifier
            argument.
        """
        args = MockArgs()
        host_id = "5ebcb1f"
        r = boom.command._delete_host_cmd(args, None, None, host_id)
        self.assertEqual(r, 0)

    def test__delete_host_cmd_no_selection(self):
        """Test the _delete_host_cmd() handler with no valid selection.
        """
        args = MockArgs()
        r = boom.command._delete_host_cmd(args, None, None, None)
        self.assertEqual(r, 1)

    def test__clone_host_cmd(self):
        """Test the _clone_host_cmd() handler with valid arguments.
        """
        args = MockArgs()
        args.host_id = "5ebcb1f"
        args.host_name = "new_host"
        r = boom.command._clone_host_cmd(args, None, None, None)
        self.assertEqual(r, 0)

    def test__show_host_cmd(self):
        """Test the _show_host_cmd() handler with valid arguments.
        """
        args = MockArgs()
        r = boom.command._show_host_cmd(args, None, None, None)
        self.assertEqual(r, 0)

    def test__show_host_cmd_with_identifier(self):
        """Test the _show_host_cmd() handler with valid arguments.
        """
        args = MockArgs()
        host_id = "1a979bb"
        r = boom.command._show_host_cmd(args, None, None, host_id)
        self.assertEqual(r, 0)

    def test__show_host_cmd_with_host_id(self):
        """Test the _show_host_cmd() handler with valid arguments.
        """
        args = MockArgs()
        args.host_id = "1a979bb"
        r = boom.command._show_host_cmd(args, None, None, None)
        self.assertEqual(r, 0)

    def test__list_host_cmd(self):
        """Test the _list_host_cmd() handler with valid arguments.
        """
        args = MockArgs()
        r = boom.command._list_host_cmd(args, None, None, None)
        self.assertEqual(r, 0)

    def test__list_host_cmd_with_identifier(self):
        """Test the _list_host_cmd() handler with valid arguments.
        """
        args = MockArgs()
        host_id = "1a979bb"
        r = boom.command._list_host_cmd(args, None, None, host_id)
        self.assertEqual(r, 0)

    def test__list_host_cmd_with_host_id(self):
        """Test the _list_host_cmd() handler with valid arguments.
        """
        args = MockArgs()
        args.host_id = "1a979bb"
        r = boom.command._list_host_cmd(args, None, None, None)
        self.assertEqual(r, 0)

    def test__list_host_cmd_with_options(self):
        """Test the _list_host_cmd() handler with valid --host-id
            argument and report control options.
        """
        args = MockArgs()
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._list_host_cmd(args, None, opts, None)
        self.assertEqual(r, 0)

    def test__list_host_cmd_with_verbose(self):
        """Test the _list_host_cmd() handler with valid --host-id
            argument and verbosity.
        """
        args = MockArgs()
        args.verbose = 1
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._list_host_cmd(args, None, opts, None)
        self.assertEqual(r, 0)

    def test__list_host_cmd_with_fields(self):
        """Test the _list_host_cmd() handler with valid --host-id
            argument and custom report field options.
        """
        args = MockArgs()
        args.options = "hostid,hostname"
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._list_host_cmd(args, None, opts, None)
        self.assertEqual(r, 0)

    def test__edit_host_cmd(self):
        """Test the _edit_host_cmd() handler with valid arguments.
        """
        args = MockArgs()
        args.host_id = "1a979bb"
        args.host_name = "notlocalhost"
        r = boom.command._edit_host_cmd(args, None, None, None)
        self.assertEqual(r, 0)

    def test__edit_host_cmd_with_invalid_options(self):
        """Test the _edit_host_cmd() handler with valid arguments.
        """
        args = MockArgs()
        args.options = "bad,touch,ricky,bad,touch"
        r = boom.command._edit_host_cmd(args, None, None, None)
        self.assertEqual(r, 1)

    def test__edit_host_cmd_with_identifier(self):
        """Test the _edit_host_cmd() handler with valid arguments.
        """
        args = MockArgs()
        args.host_name = "notlocalhost"
        host_id = "1a979bb"
        r = boom.command._edit_host_cmd(args, None, None, host_id)
        self.assertEqual(r, 0)

    def test__list_cache_cmd(self):
        args = MockArgs()
        r = boom.command._list_cache_cmd(args, None, None, None)

    def test__list_cache_cmd_with_sort_num(self):
        args = MockArgs()
        args.sort = "count"
        opts = boom.command._report_opts_from_args(args)
        r = boom.command._list_cache_cmd(args, Selection(), opts, None)
        self.assertEqual(r, 0)

    def test__show_cache_cmd(self):
        args = MockArgs()
        r = boom.command._show_cache_cmd(args, None, None, None)

    def test_boom_main_noargs(self):
        args = ['bin/boom', '--help']
        boom.command.main(args)

    def test_boom_main_list(self):
        args = ['bin/boom', 'entry', 'list']
        boom.command.main(args)

# vim: set et ts=4 sw=4 :



#!/usr/bin/env python
from setuptools import setup

from boom import __version__ as boom_version

setup(
    name='boom',
    version=boom_version,
    description=("""The Boom Boot Manager."""),
    author='Bryn M. Reeves',
    author_email='bmr@redhat.com',
    url='https://github.com/snapshotmanager/boom',
    license="GPLv2",
    test_suite="tests",
    scripts=['bin/boom'],
    packages=['boom'],
)


# vim: set et ts=4 sw=4 :



# Copyright (C) 2017 Red Hat, Inc., Bryn M. Reeves <bmr@redhat.com>
#
# hostprofile.py - Boom host profiles
#
# This file is part of the boom project.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions
# of the GNU General Public License v.2.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
"""The ``boom.hostprofile``  module defines the `HostProfile` class that
represents a host system profile. A `HostProfile` defines the identity
of a host and includes template values that override the corresponding
``OsProfile`` defaults for the respective host.

Functions are provided to read and write host system profiles from
an on-disk store, and to retrieve ``HostProfile`` instances using
various selection criteria.

The ``HostProfile`` class includes named properties for each profile
attribute ("profile key"). In addition, the class serves as a container
type, allowing attributes to be accessed via dictionary-style indexing.
This simplifies iteration over a profile's key / value pairs and allows
straightforward access to all members in scripts and the Python shell.

The keys used to access ``HostProfile`` members (and their corresponding
property names) are identical to those used by the ``OsProfile`` class.
"""
from __future__ import print_function

from boom import *
from boom.osprofile import *

from hashlib import sha1
from os.path import join as path_join
import logging
import string

# Module logging configuration
_log = logging.getLogger(__name__)
_log.set_debug_mask(BOOM_DEBUG_PROFILE)

_log_debug = _log.debug
_log_debug_profile = _log.debug_masked
_log_info = _log.info
_log_warn = _log.warning
_log_error = _log.error

#: Global host profile list
_host_profiles = []
_host_profiles_by_id = {}
_host_profiles_by_host_id = {}

#: Whether profiles have been read from disk
_host_profiles_loaded = False

#: Boom profiles directory name.
BOOM_HOST_PROFILES = "hosts"

#: File name format for Boom profiles.
BOOM_HOST_PROFILE_FORMAT = "%s-%s.host"

#: The file mode with which to create Boom profiles.
BOOM_HOST_PROFILE_MODE = 0o644

# Constants for Boom profile keys
#: Constant for the Boom host identifier profile key.
BOOM_HOST_ID = "BOOM_HOST_ID"
#: Constant for the Boom host name profile key.
BOOM_HOST_NAME = "BOOM_HOST_NAME"
#: Constant for the Boom host add options key.
BOOM_HOST_ADD_OPTS = "BOOM_HOST_ADD_OPTS"
#: Constant for the Boom host del options key.
BOOM_HOST_DEL_OPTS = "BOOM_HOST_DEL_OPTS"
#: Constant for the Boom host label key.
BOOM_HOST_LABEL = "BOOM_HOST_LABEL"

#: Constant for shared machine_id key
BOOM_ENTRY_MACHINE_ID = "BOOM_ENTRY_MACHINE_ID"

#: Ordered list of possible host profile keys, partitioned into
#: mandatory keys, optional host profile keys, keys mapping to
#: embedded ``OsProfile`` identity data, and ``OsProfile`` pattern
#: keys that may be overridden in the ``HostProfile``.
HOST_PROFILE_KEYS = [
    # HostProfile identifier
    BOOM_HOST_ID,
    # Machine hostname
    BOOM_HOST_NAME,
    # Binding to label, machine_id and OsProfile
    BOOM_ENTRY_MACHINE_ID, BOOM_OS_ID,
    # Optional host profile keys
    BOOM_HOST_LABEL, BOOM_HOST_ADD_OPTS, BOOM_HOST_DEL_OPTS,
    # Keys 7-10 OS identity keys mapped to the embedded OsProfile.
    BOOM_OS_NAME, BOOM_OS_SHORT_NAME, BOOM_OS_VERSION, BOOM_OS_VERSION_ID,
    # Keys 11-15 (OsProfile patterns) may be overridden in the host profile.
    BOOM_OS_UNAME_PATTERN, BOOM_OS_KERNEL_PATTERN, BOOM_OS_INITRAMFS_PATTERN,
    BOOM_OS_ROOT_OPTS_LVM2, BOOM_OS_ROOT_OPTS_BTRFS,
    BOOM_OS_OPTIONS
]

#: A map of Boom host profile keys to human readable key names suitable
#: for use in formatted output. These key names are used to format a
#: ``HostProfile`` object as a human readable string.
HOST_KEY_NAMES = {
    # Keys unique to HostProfile
    BOOM_HOST_ID: "Host ID",
    BOOM_HOST_NAME: "Host name",
    BOOM_HOST_LABEL: "Host label",
    BOOM_HOST_ADD_OPTS: "Add options",
    BOOM_HOST_DEL_OPTS: "Del options",
    # Keys shared with BootEntry
    BOOM_ENTRY_MACHINE_ID: "Machine ID",
    # Keys incorporated from OsProfile
    BOOM_OS_ID: OS_KEY_NAMES[BOOM_OS_ID],
    BOOM_OS_NAME: OS_KEY_NAMES[BOOM_OS_NAME],
    BOOM_OS_SHORT_NAME: OS_KEY_NAMES[BOOM_OS_SHORT_NAME],
    BOOM_OS_VERSION: OS_KEY_NAMES[BOOM_OS_VERSION],
    BOOM_OS_VERSION_ID: OS_KEY_NAMES[BOOM_OS_VERSION_ID],
    BOOM_OS_UNAME_PATTERN: OS_KEY_NAMES[BOOM_OS_UNAME_PATTERN],
    BOOM_OS_KERNEL_PATTERN: OS_KEY_NAMES[BOOM_OS_KERNEL_PATTERN],
    BOOM_OS_INITRAMFS_PATTERN: OS_KEY_NAMES[BOOM_OS_INITRAMFS_PATTERN],
    BOOM_OS_ROOT_OPTS_LVM2: OS_KEY_NAMES[BOOM_OS_ROOT_OPTS_LVM2],
    BOOM_OS_ROOT_OPTS_BTRFS: OS_KEY_NAMES[BOOM_OS_ROOT_OPTS_BTRFS],
    BOOM_OS_OPTIONS: OS_KEY_NAMES[BOOM_OS_OPTIONS]
}

#: Boom host profile keys that must exist in a valid profile.
HOST_REQUIRED_KEYS = HOST_PROFILE_KEYS[0:4]

#: Boom optional host profile configuration keys.
HOST_OPTIONAL_KEYS = HOST_PROFILE_KEYS[4:]


def _host_exists(host_id):
    """Test whether the specified ``host_id`` already exists.

        Used during ``HostProfile`` initialisation to test if the new
        ``host_id`` is already known (and to avoid passing through
        find_profiles(), which may trigger recursive profile loading).

        :param host_id: the host identifier to check for

        :returns: ``True`` if the identifier is known or ``False``
                  otherwise.
        :rtype: bool
    """
    global _host_profiles_by_host_id
    if not _host_profiles_by_host_id:
        return False
    if host_id in _host_profiles_by_host_id:
        return True
    return False


def boom_host_profiles_path():
    """Return the path to the boom host profiles directory.

        :returns: The boom host profiles path.
        :rtype: str
    """
    return path_join(get_boom_path(), BOOM_HOST_PROFILES)


def host_profiles_loaded():
    """Test whether profiles have been loaded from disk.

        :rtype: bool
        :returns: ``True`` if profiles are loaded in memory or ``False``
                  otherwise
    """
    return _host_profiles_loaded


def drop_host_profiles():
    """Drop all in-memory host profiles.
    """
    global _host_profiles, _host_profiles_by_id, _host_profiles_by_host_id
    global _host_profiles_loaded

    _host_profiles = []
    _host_profiles_by_id = {}
    _host_profiles_by_host_id = {}
    _host_profiles_loaded = False


def load_host_profiles():
    """Load HostProfile data from disk.

        Load the set of host profiles found at the path
        ``boom.hostprofile.boom_profiles_path()`` into the global host
        profile list.

        This function may be called to explicitly load, or reload the
        set of profiles on-disk. Profiles are also loaded implicitly
        if an API function or method that requires access to profiles
        is invoked (for example, ``boom.bootloader.load_entries()``.

        :returns: None
    """
    global _host_profiles_loaded
    drop_host_profiles()
    profiles_path = boom_host_profiles_path()
    load_profiles_for_class(HostProfile, "Host", profiles_path, "host")

    _host_profiles_loaded = True
    _log_debug("Loaded %d host profiles" % len(_host_profiles))


def write_host_profiles(force=False):
    """Write all HostProfile data to disk.

        Write the current list of host profiles to the directory located
        at ``boom.osprofile.boom_profiles_path()``.

        :rtype: None
    """
    global _host_profiles
    _log_debug("Writing host profiles to %s" % boom_host_profiles_path())
    for hp in _host_profiles:
        try:
            hp.write_profile(force)
        except Exception as e:
            _log_warn("Failed to write HostProfile(machine_id='%s'): %s" %
                      (hp.disp_machine_id, e))


def min_host_id_width():
    """Calculate the minimum unique width for host_id values.

        Calculate the minimum width to ensure uniqueness when displaying
        host_id values.

        :returns: the minimum host_id width.
        :rtype: int
    """
    return min_id_width(7, _host_profiles, "host_id")


def min_machine_id_width():
    """Calculate the minimum unique width for host_id values.

        Calculate the minimum width to ensure uniqueness when displaying
        host_id values.

        :returns: the minimum host_id width.
        :rtype: int
    """
    return min_id_width(7, _host_profiles, "machine_id")


def select_host_profile(s, hp):
    """Test the supplied host profile against selection criteria.

        Test the supplied ``HostProfile`` against the selection criteria
        in ``s`` and return ``True`` if it passes, or ``False``
        otherwise.

        :param s: The selection criteria
        :param hp: The ``HostProfile`` to test
        :rtype: bool
        :returns: True if ``hp`` passes selection or ``False`` otherwise.
    """
    if s.host_id and not hp.host_id.startswith(s.host_id):
        return False
    if s.machine_id and hp.machine_id != s.machine_id:
        return False
    if s.host_name and hp.host_name != s.host_name:
        return False
    if s.host_label and hp.label != s.host_label:
        return False
    if s.host_short_name and hp.short_name != s.host_short_name:
        return False
    if s.host_add_opts and hp.add_opts != s.host_add_opts:
        return False
    if s.host_del_opts and hp.del_opts != s.host_del_opts:
        return False
    if s.os_id and not hp.os_id.startswith(s.os_id):
        return False
    if s.os_name and hp.os_name != s.os_name:
        return False
    if s.os_short_name and hp.os_short_name != s.os_short_name:
        return False
    if s.os_version and hp.os_version != s.os_version:
        return False
    if s.os_version_id and hp.os_version_id != s.os_version_id:
        return False
    if s.os_uname_pattern and hp.uname_pattern != s.os_uname_pattern:
        return False
    if s.os_kernel_pattern and hp.kernel_pattern != s.os_kernel_pattern:
        return False
    if (s.os_initramfs_pattern and
            hp.initramfs_pattern != s.os_initramfs_pattern):
        return False
    if s.os_options and hp.options != s.os_options:
        return False
    return True


def find_host_profiles(selection=None, match_fn=select_host_profile):
    """Find host profiles matching selection criteria.

        Return a list of ``HostProfile`` objects matching the specified
        criteria. Matching proceeds as the logical 'and' of all criteria.
        Criteria that are unset (``None``) are ignored.

        If the optional ``match_fn`` parameter is specified, the match
        criteria parameters are ignored and each ``HostProfile`` is
        tested in turn by calling ``match_fn``. If the matching function
        returns ``True`` the ``HostProfile`` will be included in the
        results.

        If no ``HostProfile`` matches the specified criteria the empty
        list is returned.

        Host profiles will be automatically loaded from disk if they are
        not already in memory.

        :param selection: A ``Selection`` object specifying the match
                          criteria for the operation.
        :param match_fn: An optional match function to test profiles.
        :returns: a list of ``HostProfile`` objects.
        :rtype: list
    """
    # Use null search criteria if unspecified
    selection = selection if selection else Selection()

    selection.check_valid_selection(host=True)

    if not host_profiles_loaded():
        load_host_profiles()

    matches = []

    _log_debug_profile("Finding host profiles for %s" % repr(selection))
    for hp in _host_profiles:
        if match_fn(selection, hp):
            matches.append(hp)
    _log_debug_profile("Found %d host profiles" % len(matches))
    matches.sort(key=lambda h: h.host_name)

    return matches


def get_host_profile_by_id(machine_id, label=""):
    """Find a HostProfile by its machine_id.

        Return the HostProfile object corresponding to ``machine_id``,
        or ``None`` if it is not found.

        :rtype: HostProfile
        :returns: An HostProfile matching machine_id or None if no match
                  was found.
    """
    global _host_profiles, _host_profiles_by_id, _host_profiles_by_host_id
    if not host_profiles_loaded():
        load_host_profiles()
    if machine_id in _host_profiles_by_id:
        if label in _host_profiles_by_id[machine_id]:
            return _host_profiles_by_id[machine_id][label]
    return None


def match_host_profile(entry):
    """Attempt to match a BootEntry to a corresponding HostProfile.

        Attempt to find a loaded ``HostProfile`` object with the a
        ``machine_id`` that matches the supplied ``BootEntry``.
        Checking terminates on the first matching ``HostProfile``.

        :param entry: A ``BootEntry`` object with no attached
                      ``HostProfile``.
        :returns: The corresponding ``HostProfile`` for the supplied
                  ``BootEntry`` or ``None`` if no match is found.
        :rtype: ``BootEntry`` or ``NoneType``.
    """
    global _host_profiles, _host_profiles_loaded

    if not host_profiles_loaded():
        load_host_profiles()

    _log_debug("Attempting to match profile for BootEntry(title='%s', "
               "version='%s') with machine_id='%s'" %
               (entry.title, entry.version, entry.machine_id))

    # Attempt to match by uname pattern
    for hp in _host_profiles:
        if hp.machine_id == entry.machine_id:
            _log_debug("Matched BootEntry(version='%s', boot_id='%s') "
                       "to HostProfile(name='%s', machine_id='%s')" %
                       (entry.version, entry.disp_boot_id, hp.host_name,
                        hp.machine_id))
            return hp

    return None


class HostProfile(BoomProfile):
    """ Class HostProfile implements Boom host system profiles.

        Objects of type HostProfile define a host identiry, and optional
        fields or ``BootParams`` modifications to be applied to the
        specified host.

        Host profiles may modify any non-identity ``OsProfile`` key,
        either adding to or replacing the value defined by an embedded
        ``OsProfile`` instance.
    """
    _profile_data = None
    _unwritten = False
    _comments = None

    _profile_keys = HOST_PROFILE_KEYS
    _required_keys = HOST_REQUIRED_KEYS
    _identity_key = BOOM_HOST_ID

    _osp = None

    def _key_data(self, key):
        if key in self._profile_data:
            return self._profile_data[key]
        if key in self.osp._profile_data:
            return self.osp._profile_data[key]
        return None

    def _have_key(self, key):
        """Test for presence of a Host or Os profile key.
        """
        return key in self._profile_data or key in self.osp._profile_data

    def __str__(self):
        """Format this HostProfile as a human readable string.

            Profile attributes are printed as "Name: value, " pairs,
            with like attributes grouped together onto lines.

            :returns: A human readable string representation of this
                      HostProfile.

            :rtype: string
        """
        # FIXME HostProfile breaks
        breaks = [
            BOOM_HOST_ID, BOOM_HOST_NAME, BOOM_OS_ID, BOOM_ENTRY_MACHINE_ID,
            BOOM_HOST_LABEL, BOOM_OS_VERSION, BOOM_OS_UNAME_PATTERN,
            BOOM_HOST_DEL_OPTS, BOOM_OS_INITRAMFS_PATTERN,
            BOOM_OS_ROOT_OPTS_LVM2, BOOM_OS_ROOT_OPTS_BTRFS, BOOM_OS_OPTIONS
        ]

        fields = [f for f in HOST_PROFILE_KEYS if self._have_key(f)]
        hp_str = ""
        tail = ""
        for f in fields:
            hp_str += '%s: "%s"' % (HOST_KEY_NAMES[f], self._key_data(f))
            tail = ",\n" if f in breaks else ", "
            hp_str += tail
        hp_str = hp_str.rstrip(tail)
        return hp_str

    def __repr__(self):
        """Format this HostProfile as a machine readable string.

            Return a machine-readable representation of this ``HostProfile``
            object. The string is formatted as a call to the ``HostProfile``
            constructor, with values passed as a dictionary to the
            ``profile_data`` keyword argument.

            :returns: a string representation of this ``HostProfile``.
            :rtype: string
        """
        hp_str = "HostProfile(profile_data={"
        fields = [f for f in HOST_PROFILE_KEYS if self._have_key(f)]
        for f in fields:
            hp_str += '%s:"%s", ' % (f, self._key_data(f))
        hp_str = hp_str.rstrip(", ")
        return hp_str + "})"

    def __setitem__(self, key, value):
        """Set the specified ``HostProfile`` key to the given value.

            :param key: the ``HostProfile`` key to be set.
            :param value: the value to set for the specified key.
        """

        # FIXME: duplicated from OsProfile.__setitem__ -> factor
        # osprofile.check_format_key_value(key, value)
        # and include isstr() key name validation etc.

        # Map hp key names to a list of format keys which must not
        # appear in that key's value: e.g. %{kernel} in the kernel
        # pattern profile key.
        bad_key_map = {
            BOOM_OS_KERNEL_PATTERN: [FMT_KERNEL],
            BOOM_OS_INITRAMFS_PATTERN: [FMT_INITRAMFS],
            BOOM_OS_ROOT_OPTS_LVM2: [FMT_ROOT_OPTS],
            BOOM_OS_ROOT_OPTS_BTRFS: [FMT_ROOT_OPTS],
        }

        def _check_format_key_value(key, value, bad_keys):
            for bad_key in bad_keys:
                if bad_key in value:
                    raise ValueError("HostProfile.%s cannot contain %s"
                                     % (key, key_from_key_name(bad_key)))

        if not isinstance(key, str):
            raise TypeError("HostProfile key must be a string.")

        if key not in HOST_PROFILE_KEYS:
            raise ValueError("Invalid HostProfile key: %s" % key)

        if key in bad_key_map:
            _check_format_key_value(key, value, bad_key_map[key])

        self._profile_data[key] = value

    def _generate_id(self):
        """Generate a new host identifier.

            Generate a new sha1 profile identifier for this profile,
            using the name, machine_id, and os_id and store it in
            _profile_data.

            :returns: None
        """
        hashdata = (self.machine_id + self.label)

        digest = sha1(hashdata.encode('utf-8')).hexdigest()
        self._profile_data[BOOM_HOST_ID] = digest

    def __set_os_profile(self):
        """Set this ``HostProfile``'s ``osp`` member to the
            corresponding profile for the set ``os_id``.
        """
        os_id = self._profile_data[BOOM_OS_ID]
        osps = find_profiles(Selection(os_id=os_id))
        if not osps:
            raise ValueError("OsProfile not found: %s" % os_id)
        if len(osps) > 1:
            raise ValueError("OsProfile identifier '%s' is ambiguous" % os_id)

        self.osp = osps[0]

    def _append_profile(self):
        """Append a HostProfile to the global profile list
        """
        global _host_profiles, _host_profiles_by_id, _host_profiles_by_host_id
        if _host_exists(self.host_id):
            raise ValueError("Profile already exists (host_id=%s)" %
                             self.disp_host_id)

        _host_profiles.append(self)
        machine_id = self.machine_id
        if machine_id not in _host_profiles_by_id:
            _host_profiles_by_id[machine_id] = {}
        _host_profiles_by_id[machine_id][self.label] = self
        _host_profiles_by_host_id[self.host_id] = self

    def _from_data(self, host_data, dirty=True):
        """Initialise a ``HostProfile`` from in-memory data.

            Initialise a new ``HostProfile`` object using the profile
            data in the `host_data` dictionary.

            This method should not be called directly: to build a new
            `Hostprofile`` object from in-memory data, use the class
            initialiser with the ``host_data`` argument.

            :returns: None
        """
        err_str = "Invalid profile data (missing %s)"

        for key in HOST_REQUIRED_KEYS:
            if key == BOOM_HOST_ID:
                continue
            if key not in host_data:
                raise ValueError(err_str % key)

        self._profile_data = dict(host_data)

        if BOOM_HOST_ID not in self._profile_data:
            self._generate_id()

        self.__set_os_profile()

        if dirty:
            self._dirty()

        self._append_profile()

    def __init__(self, machine_id=None, host_name=None, label=None, os_id=None,
                 kernel_pattern=None, initramfs_pattern=None,
                 root_opts_lvm2=None, root_opts_btrfs=None,
                 add_opts="", del_opts="",
                 options=None, profile_file=None, profile_data=None):
        """Initialise a new ``HostProfile`` object.

            If neither ``profile_file`` nor ``profile_data`` is given,
            all of ``machine_id``, ``name``, and ``os_id`` must be given.

            These values form the host profile identity and are used to
            generate the profile unique identifier.

            :param host_name: The hostname of this system
            :param os_id: An OS identifier specifying the ``OsProfile``
                          to use with this host profile.
            :param profile_data: An optional dictionary mapping from
                                 ``BOOM_*`` keys to profile values.
            :param profile_file: An optional path to a file from which
                                 profile data should be loaded. The file
                                 should be in Boom host profile format,
                                 with ``BOOM_*`` key=value pairs.
            :returns: A new ``HostProfile`` object.
            :rtype: class HostProfile
        """
        global _host_profiles
        self._profile_data = {}

        # Initialise BoomProfile base class
        super(HostProfile, self).__init__(HOST_PROFILE_KEYS,
                                          HOST_REQUIRED_KEYS, BOOM_HOST_ID)

        if profile_data and profile_file:
            raise ValueError("Only one of 'profile_data' or 'profile_file' "
                             "may be specified.")

        if profile_data:
            self._from_data(profile_data)
            return
        if profile_file:
            self._from_file(profile_file)
            return

        self._dirty()

        required_args = [machine_id, host_name, os_id]
        if any([not val for val in required_args]):
            raise ValueError("Invalid host profile arguments: machine_id, "
                             "host_name, and os_id are mandatory.")

        osps = find_profiles(Selection(os_id=os_id))
        if not osps:
            raise ValueError("No matching profile found for os_id=%s" % os_id)
        if len(osps) > 1:
            raise ValueError("OsProfile ID is ambiguous: %s" % os_id)
        os_id = osps[0].os_id

        self._profile_data[BOOM_ENTRY_MACHINE_ID] = machine_id
        self._profile_data[BOOM_HOST_NAME] = host_name
        self._profile_data[BOOM_OS_ID] = os_id
        self._profile_data[BOOM_HOST_LABEL] = label

        # Only set keys that have a value in the host profile data dict
        if kernel_pattern:
            self._profile_data[BOOM_OS_KERNEL_PATTERN] = kernel_pattern
        if initramfs_pattern:
            self._profile_data[BOOM_OS_INITRAMFS_PATTERN] = initramfs_pattern
        if root_opts_lvm2:
            self._profile_data[BOOM_OS_ROOT_OPTS_LVM2] = root_opts_lvm2
        if root_opts_btrfs:
            self._profile_data[BOOM_OS_ROOT_OPTS_BTRFS] = root_opts_btrfs
        if add_opts:
            self._profile_data[BOOM_HOST_ADD_OPTS] = add_opts
        if del_opts:
            self._profile_data[BOOM_HOST_DEL_OPTS] = del_opts
        if options:
            self._profile_data[BOOM_OS_OPTIONS] = options

        self.__set_os_profile()

        self._generate_id()
        _host_profiles.append(self)

    # We use properties for the HostProfile attributes: this is to
    # allow the values to be stored in a dictionary. Although
    # properties are quite verbose this reduces the code volume
    # and complexity needed to marshal and unmarshal the various
    # file formats used, as well as conversion to and from string
    # representations of HostProfile objects.

    # Keys obtained from os-release data form the profile's identity:
    # the corresponding attributes are read-only.

    # HostProfile properties:
    #
    #  Profile identity properties (ro):
    #   host_id
    #   disp_os_id
    #
    #  Profile identity properties (rw):
    #   host_name
    #   machine_id
    #   os_id
    #
    #  Properties mapped to OsProfile (ro)
    #   os_name
    #   os_short_name
    #   os_version
    #   os_version_id
    #   uname_pattern
    #
    # Properties overridden or mapped to OsProfile (rw)
    #   kernel_pattern
    #   initramfs_pattern
    #   root_opts_lvm2
    #   root_opts_btrfs
    #   options
    #
    # HostProfile specific properties (rw)
    #   label
    #   add_opts
    #   del_opts

    @property
    def disp_os_id(self):
        """The display os_id of this profile.

            Return the shortest prefix of this OsProfile's os_id that
            is unique within the current set of loaded profiles.

            :getter: return this OsProfile's os_id.
            :type: str
        """
        return self.osp.disp_os_id

    @property
    def host_id(self):
        if BOOM_HOST_ID not in self._profile_data:
            self._generate_id()
        return self._profile_data[BOOM_HOST_ID]

    @property
    def disp_host_id(self):
        """The display host_id of this profile

            Return the shortest prefix of this HostProfile's os_id that
            is unique within the current set of loaded profiles.

            :getter: return this HostProfile's display host_id.
            :type: str
        """
        return self.host_id[:min_host_id_width()]

    @property
    def disp_machine_id(self):
        """The machine_id of this host profile.
            Return the shortest prefix of this HostProfile's os_id that
            is unique within the current set of loaded profiles.

            :getter: return this HostProfile's display host_id.
            :type: str
        """
        return self.machine_id[:min_machine_id_width()]

    @property
    def machine_id(self):
        """The machine_id of this host profile.
            Return the shortest prefix of this HostProfile's os_id that
            is unique within the current set of loaded profiles.

            :getter: return this ``HostProfile``'s display host_id.
            :setter: change this ``HostProfile``'s ``machine_id``. This
                     will change the ``host_id``.
            :type: str
        """
        return self._profile_data[BOOM_ENTRY_MACHINE_ID]

    @machine_id.setter
    def machine_id(self, value):
        if value == self._profile_data[BOOM_ENTRY_MACHINE_ID]:
            return
        self._profile_data[BOOM_ENTRY_MACHINE_ID] = value
        self._dirty()
        self._generate_id()

    @property
    def os_id(self):
        """The ``os_id`` of this profile.

            :getter: returns the ``os_id`` as a string.
            :type: string
        """
        return self.osp.os_id

    @os_id.setter
    def os_id(self, value):
        if value == self._profile_data[BOOM_OS_ID]:
            return
        self._profile_data[BOOM_OS_ID] = value
        self.__set_os_profile()
        self._dirty()
        self._generate_id()

    @property
    def osp(self):
        """The ``OsProfile`` used by this ``HostProfile``.

            :getter: returns the ``OsProfile`` object used by this
                     ``HostProfile``.
            :setter: stores a new ``OsProfile`` for use by this
                     ``HostProfile`` and updates the stored ``os_id``
                     value in the host profile.
        """
        return self._osp

    @osp.setter
    def osp(self, osp):
        if self._osp and osp.os_id == self._osp.os_id:
            return
        self._osp = osp
        self._profile_data[BOOM_OS_ID] = osp.os_id
        self._dirty()
        self._generate_id()

    @property
    def host_name(self):
        """The ``host_name`` of this profile.

            Normally set to the hostname of the system corresponding to
            this ``HostProfile``.

            :getter: returns the ``host_name`` as a string.
            :type: string
        """
        return self._profile_data[BOOM_HOST_NAME]

    @host_name.setter
    def host_name(self, value):
        if value == self._profile_data[BOOM_HOST_NAME]:
            return
        self._profile_data[BOOM_HOST_NAME] = value
        self._dirty()
        self._generate_id()

    @property
    def short_name(self):
        """The ``short_name`` of this profile.

            If ``HostProfile.host_name`` appears to contain a DNS-style name,
            return only the host portion.

            :getter: returns the ``short_name`` as a string.
            :type: string
        """
        host_name = self._profile_data[BOOM_HOST_NAME]
        return host_name.split(".")[0] if "." in host_name else host_name

    #
    #  Properties mapped to OsProfile
    #

    @property
    def os_name(self):
        """The ``os_name`` of this profile.

            :getter: returns the ``os_name`` as a string.
            :type: string
        """
        return self.osp.os_name

    @property
    def os_short_name(self):
        """The ``os_short_name`` of this profile.

            :getter: returns the ``os_short_name`` as a string.
            :type: string
        """
        return self.osp.os_short_name

    @property
    def os_version(self):
        """The ``os_version`` of this profile.

            :getter: returns the ``os_version`` as a string.
            :type: string
        """
        return self.osp.os_version

    @property
    def os_version_id(self):
        """The ``os_version_id`` of this profile.

            :getter: returns the ``os_version_id`` as a string.
            :type: string
        """
        return self.osp.os_version_id

    @property
    def uname_pattern(self):
        """The current ``uname_pattern`` setting of this profile.

            :getter: returns the ``uname_pattern`` as a string.
            :setter: stores a new ``uname_pattern`` setting.
            :type: string
        """
        if BOOM_OS_UNAME_PATTERN in self._profile_data:
            return self._profile_data[BOOM_OS_UNAME_PATTERN]
        return self.osp.uname_pattern

    #
    # Properties overridden or mapped to OsProfile
    #

    @property
    def kernel_pattern(self):
        """The current ``kernel_pattern`` setting of this profile.

            :getter: returns the ``kernel_pattern`` as a string.
            :setter: stores a new ``kernel_pattern`` setting.
            :type: string
        """
        if BOOM_OS_KERNEL_PATTERN in self._profile_data:
            return self._profile_data[BOOM_OS_KERNEL_PATTERN]
        return self.osp.kernel_pattern

    @kernel_pattern.setter
    def kernel_pattern(self, value):
        kernel_key = key_from_key_name(FMT_KERNEL)
        if kernel_key in value:
            raise ValueError("HostProfile.kernel cannot contain %s" %
                             kernel_key)
        self._profile_data[BOOM_OS_KERNEL_PATTERN] = value
        self._dirty()

    @property
    def initramfs_pattern(self):
        """The current ``initramfs_pattern`` setting of this profile.

            :getter: returns the ``initramfs_pattern`` as a string.
            :setter: store a new ``initramfs_pattern`` setting.
            :type: string
        """
        if BOOM_OS_INITRAMFS_PATTERN in self._profile_data:
            return self._profile_data[BOOM_OS_INITRAMFS_PATTERN]
        return self.osp.initramfs_pattern

    @initramfs_pattern.setter
    def initramfs_pattern(self, value):
        initramfs_key = key_from_key_name(FMT_INITRAMFS)
        if initramfs_key in value:
            raise ValueError("HostProfile.initramfs cannot contain %s" %
                             initramfs_key)
        self._profile_data[BOOM_OS_INITRAMFS_PATTERN] = value
        self._dirty()

    @property
    def root_opts_lvm2(self):
        """The current LVM2 root options setting of this profile.

            :getter: returns the ``root_opts_lvm2`` value as a string.
            :setter: store a new ``root_opts_lvm2`` value.
            :type: string
        """
        if BOOM_OS_ROOT_OPTS_LVM2 in self._profile_data:
            return self._profile_data[BOOM_OS_ROOT_OPTS_LVM2]

        return self.osp.root_opts_lvm2

    @root_opts_lvm2.setter
    def root_opts_lvm2(self, value):
        root_opts_key = key_from_key_name(FMT_ROOT_OPTS)
        if root_opts_key in value:
            raise ValueError("HostProfile.root_opts_lvm2 cannot contain ""%s" %
                             root_opts_key)
        self._profile_data[BOOM_OS_ROOT_OPTS_LVM2] = value
        self._dirty()

    @property
    def root_opts_btrfs(self):
        """The current BTRFS root options setting of this profile.

            :getter: returns the ``root_opts_btrfs`` value as a string.
            :setter: store a new ``root_opts_btrfs`` value.
            :type: string
        """
        if BOOM_OS_ROOT_OPTS_BTRFS in self._profile_data:
            return self._profile_data[BOOM_OS_ROOT_OPTS_BTRFS]
        return self.osp.root_opts_btrfs

    @root_opts_btrfs.setter
    def root_opts_btrfs(self, value):
        root_opts_key = key_from_key_name(FMT_ROOT_OPTS)
        if root_opts_key in value:
            raise ValueError("HostProfile.root_opts_btrfs cannot contain %s" %
                             root_opts_key)
        self._profile_data[BOOM_OS_ROOT_OPTS_BTRFS] = value
        self._dirty()

    @property
    def options(self):
        """The current kernel command line options setting for this
            profile.

            :getter: returns the ``options`` value as a string.
            :setter: store a new ``options`` value.
            :type: string
        """
        if BOOM_OS_OPTIONS in self._profile_data:
            return self._profile_data[BOOM_OS_OPTIONS]
        return self.osp.options

    @options.setter
    def options(self, value):
        self._profile_data[BOOM_OS_OPTIONS] = value
        self._dirty()

    @property
    def title(self):
        """The current title template for this profile.

            :getter: returns the ``title`` value as a string.
            :setter: store a new ``title`` value.
            :type: string
        """
        if BOOM_OS_TITLE not in self._profile_data:
            return None
        return self._profile_data[BOOM_OS_TITLE]

    @title.setter
    def title(self, value):
        if not value:
            # It is valid to set an empty title in a HostProfile as long
            # as the OsProfile defines one.
            if not self.osp or not self.osp.title:
                raise ValueError("Entry title cannot be empty")
        self._profile_data[BOOM_OS_TITLE] = value
        self._dirty()

    @property
    def optional_keys(self):
        if not self.osp or not self.osp.optional_keys:
            return ""
        return self.osp.optional_keys

    #
    # HostProfile specific properties
    #

    @property
    def add_opts(self):
        if BOOM_HOST_ADD_OPTS in self._profile_data:
            return self._profile_data[BOOM_HOST_ADD_OPTS]
        return ""

    @add_opts.setter
    def add_opts(self, opts):
        self._profile_data[BOOM_HOST_ADD_OPTS] = opts
        self._dirty()

    @property
    def del_opts(self):
        if BOOM_HOST_DEL_OPTS in self._profile_data:
            return self._profile_data[BOOM_HOST_DEL_OPTS]
        return ""

    @del_opts.setter
    def del_opts(self, opts):
        self._profile_data[BOOM_HOST_DEL_OPTS] = opts
        self._dirty()

    @property
    def label(self):
        if BOOM_HOST_LABEL in self._profile_data:
            return self._profile_data[BOOM_HOST_LABEL]
        return ""

    @label.setter
    def label(self, value):
        valid_chars = string.ascii_letters + string.digits + "_- "

        if BOOM_HOST_LABEL in self._profile_data:
            if self._profile_data[BOOM_HOST_LABEL] == value:
                return

        for c in value:
            if c not in valid_chars:
                raise ValueError("Invalid host label character: '%s'" % c)

        self._profile_data[BOOM_HOST_LABEL] = value
        self._dirty()
        self._generate_id()

    def _profile_path(self):
        """Return the path to this profile's on-disk data.

            Return the full path to this HostProfile in the Boom profiles
            directory (or the location to which it will be written, if
            it has not yet been written).

            :rtype: str
            :returns: The absolute path for this HostProfile's file
        """
        if self.label:
            label = self.label
            if " " in label:
                label = label.replace(" ", "_")
            names = (self.short_name, label)
            name_fmt = "%s-%s"
        else:
            names = (self.short_name)
            name_fmt = "%s"
        profile_name = name_fmt % names
        profile_id = (self.host_id, profile_name)
        profile_path_name = BOOM_HOST_PROFILE_FORMAT % profile_id
        return path_join(boom_host_profiles_path(), profile_path_name)

    def write_profile(self, force=False):
        """Write out profile data to disk.

            Write out this ``HostProfile``'s data to a file in Boom
            format to the paths specified by the current configuration.

            Currently the ``machine_id`` and ``name`` keys are used to
            construct the file name.

            If the value of ``force`` is ``False`` and the ``HostProfile``
            is not currently marked as dirty (either new, or modified
            since the last load operation) the write will be skipped.

            :param force: Force this profile to be written to disk even
                          if the entry is unmodified.
            :raises: ``OsError`` if the temporary entry file cannot be
                     renamed, or if setting file permissions on the
                     new entry file fails.
        """
        path = boom_host_profiles_path()
        mode = BOOM_HOST_PROFILE_MODE
        self._write_profile(self.host_id, path, mode, force=force)

    def delete_profile(self):
        """Delete on-disk data for this profile.

            Remove the on-disk profile corresponding to this
            ``HostProfile`` object. This will permanently erase the
            current file (although the current data may be re-written at
            any time by calling ``write_profile()`` before the object is
            disposed of).

            :rtype: ``NoneType``
            :raises: ``OsError`` if an error occurs removing the file or
                     ``ValueError`` if the profile does not exist.
        """
        global _host_profiles, _host_profiles_by_id, _host_profiles_by_host_id
        self._delete_profile(self.host_id)

        machine_id = self.machine_id
        host_id = self.host_id
        if _host_profiles and self in _host_profiles:
            _host_profiles.remove(self)
        if _host_profiles_by_id and machine_id in _host_profiles_by_id:
            _host_profiles_by_id.pop(machine_id)
        if _host_profiles_by_host_id and host_id in _host_profiles_by_host_id:
            _host_profiles_by_host_id.pop(host_id)


__all__ = [
    # Host profiles
    'HostProfile',
    'drop_host_profiles', 'load_host_profiles', 'write_host_profiles',
    'host_profiles_loaded', 'find_host_profiles', 'select_host_profile',
    'get_host_profile_by_id', 'match_host_profile', 'select_host_profile',

    # Host profile keys
    'BOOM_HOST_ID', 'BOOM_HOST_NAME',
    'BOOM_HOST_ADD_OPTS', 'BOOM_HOST_DEL_OPTS', 'BOOM_HOST_LABEL',
    'HOST_PROFILE_KEYS', 'HOST_REQUIRED_KEYS', 'HOST_OPTIONAL_KEYS',

    # Path configuration
    'boom_host_profiles_path',
]

# vim: set et ts=4 sw=4 :
